import re
from abc import ABC, abstractmethod
from collections.abc import MutableSequence

import numpy as np
from numba import float32, float64, njit
from scipy.optimize import LinearConstraint


class ParametricConstraint(ABC):
    def __init__(self, binding_tol=1e-4):
        self.binding_tol = binding_tol

    def is_binding(self, x):
        return np.absolute(self.fun(x)) < self.binding_tol

    @abstractmethod
    def get_parameters(self):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def get_binding_note(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def fun(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def jac(self, x):
        raise NotImplementedError("abstract base class, use a derived class instead")

    @abstractmethod
    def link_model(self, model, scale=None):
        raise NotImplementedError("abstract base class, use a derived class instead")

    def as_constraint_dicts(self):
        return [
            dict(type="ineq", fun=self.fun, jac=self.jac),
        ]


class RatioBound(ParametricConstraint):
    def __init__(
        self,
        p_num,
        p_den,
        min_ratio=None,
        max_ratio=None,
        model=None,
        scale=10,
        binding_tol=1e-4,
    ):
        self.i_num = 0
        self.i_den = 0
        self.cmin_num = 0.0
        self.cmin_den = 0.0
        self.cmax_num = 0.0
        self.cmax_den = 0.0
        super().__init__(binding_tol=binding_tol)
        self.cmax_num = 0
        self.cmax_den = 0
        self.p_num = p_num
        self.p_den = p_den
        self.min_ratio = min_ratio
        self.max_ratio = max_ratio
        self.scale = 1
        self.link_model(model, scale)

    def __repr__(self):
        return f"larch.RatioBound({self.p_num},{self.p_den},{self.min_ratio},{self.max_ratio})"

    def link_model(self, model, scale=None):
        if scale is not None:
            self.scale = scale
        if model is not None:
            self.i_num = model.get_param_loc(self.p_num)
            self.i_den = model.get_param_loc(self.p_den)
            self.len = model.n_params
            if self.min_ratio is not None:
                scaling = max(self.min_ratio, 1 / self.min_ratio) * self.scale
                if model.get_value(self.p_den, kind="minimum") >= 0:
                    # positive denominator
                    self.cmin_num = 1 * scaling
                    self.cmin_den = -self.min_ratio * scaling
                elif model.get_value(self.p_den, kind="maximum") <= 0:
                    # negative denominator
                    self.cmin_num = -1 * scaling
                    self.cmin_den = self.min_ratio * scaling
                else:
                    raise ValueError(
                        "denominator must be bounded to be non-positive or non-negative"
                    )
            else:
                self.cmin_num = 0
                self.cmin_den = 0
            if self.max_ratio is not None:
                scaling = max(self.max_ratio, 1 / self.max_ratio) * self.scale
                if model.get_value(self.p_den, kind="minimum") >= 0:
                    # positive denominator
                    self.cmax_num = -1 * scaling
                    self.cmax_den = self.max_ratio * scaling
                elif model.get_value(self.p_den, kind="maximum") <= 0:
                    # negative denominator
                    self.cmax_num = 1 * scaling
                    self.cmax_den = -self.max_ratio * scaling
                else:
                    raise ValueError(
                        "denominator must be bounded to be non-positive or non-negative"
                    )
            else:
                self.cmax_num = 0
                self.cmax_den = 0

    def _min_fun(self, x):
        return x[self.i_num] * self.cmin_num + x[self.i_den] * self.cmin_den

    def _max_fun(self, x):
        return x[self.i_num] * self.cmax_num + x[self.i_den] * self.cmax_den

    def fun(self, x):
        return min(
            self._min_fun(x),
            self._max_fun(x),
        )

    def jac(self, x):
        j = np.zeros_like(x)
        if self._min_fun(x) < self._max_fun(x):
            j[self.i_num] = self.cmin_num
            j[self.i_den] = self.cmin_den
        else:
            j[self.i_num] = self.cmax_num
            j[self.i_den] = self.cmax_den
        return j

    def get_parameters(self):
        return (self.p_num, self.p_den)

    def get_binding_note(self, x):
        if self._min_fun(x) < self._max_fun(x):
            return f"{self.p_num!s} / {self.p_den!s} ≥ {self.min_ratio}"
        else:
            return f"{self.p_num!s} / {self.p_den!s} ≤ {self.max_ratio}"

    def as_linear_constraints(self):
        a = np.zeros([2, self.len], dtype="float64")
        a[0, self.i_num] = self.cmin_num
        a[0, self.i_den] = self.cmin_den
        a[1, self.i_num] = self.cmax_num
        a[1, self.i_den] = self.cmax_den
        return [LinearConstraint(a, 0, np.inf)]

    def as_soft_penalty(self):
        from .numbamodel import d_softplus, softplus  # here to avoid circular imports

        i_num = self.i_num
        i_den = self.i_den
        cmin_num = self.cmin_num
        cmin_den = self.cmin_den
        cmax_num = self.cmax_num
        cmax_den = self.cmax_den
        scale = self.scale

        @njit(
            [
                float32(float32[:], float32, float32),
                float64(float64[:], float64, float64),
            ]
        )
        def penalty(x, intensity, sharpness=1.0):
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            return -softplus(-np.minimum(_min, _max) * scale * intensity, sharpness)

        @njit(
            [
                float32[:](float32[:], float32, float32),
                float64[:](float64[:], float64, float64),
            ]
        )
        def dpenalty(x, intensity, sharpness=1.0):
            j = np.zeros_like(x)
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            partial = (
                d_softplus(-np.minimum(_min, _max), sharpness * scale * intensity)
                * scale
                * intensity
            )
            if _min < _max:
                j[i_num] = cmin_num * partial
                j[i_den] = cmin_den * partial
            else:
                j[i_num] = cmax_num * partial
                j[i_den] = cmax_den * partial
            return j

        @njit(
            [
                float32[:](float32[:], float32),
                float64[:](float64[:], float64),
            ]
        )
        def dpenalty_money(x, intensity):
            j = np.zeros_like(x)
            _min = x[i_num] * cmin_num + x[i_den] * cmin_den
            _max = x[i_num] * cmax_num + x[i_den] * cmax_den
            if np.absolute(_min) < 1e-5:
                partial = 0.5 * scale * intensity
                j[i_num] = cmin_num * partial
                j[i_den] = cmin_den * partial
            elif np.absolute(_max) < 1e-5:
                partial = 0.5 * scale * intensity
                j[i_num] = cmax_num * partial
                j[i_den] = cmax_den * partial
            return j

        return penalty, dpenalty, dpenalty_money

    def __eq__(self, other):
        if not isinstance(other, RatioBound):
            return False
        if (
            (self.p_num == other.p_num)
            and (self.p_den == other.p_den)
            and (self.min_ratio == other.min_ratio)
            and (self.max_ratio == other.max_ratio)
        ):
            return True
        return False


class OrderingBound(ParametricConstraint):
    def __init__(self, p_less, p_more=None, model=None, scale=10, binding_tol=1e-4):
        # default initializers
        self.i_less = 0
        self.i_more = 0
        if p_more is None:
            if "<=" in p_less:
                p_less, p_more = (_.strip() for _ in p_less.split("<=", 1))
            elif "<" in p_less:
                p_less, p_more = (_.strip() for _ in p_less.split("<", 1))
            elif ">=" in p_less:
                p_more, p_less = (_.strip() for _ in p_less.split(">=", 1))
            elif ">" in p_less:
                p_more, p_less = (_.strip() for _ in p_less.split(">", 1))
            else:
                raise ValueError(f"cannot interpret {p_less} as ordering")
        super().__init__(binding_tol=binding_tol)
        self.p_less = p_less
        self.p_more = p_more
        self.scale = 1
        if model is not None:
            self.len = model.n_params
            self.link_model(model, scale)
        else:
            self.len = 0

    def __repr__(self):
        return f"larch.OrderingBound({self.p_less},{self.p_more})"

    def link_model(self, model, scale=None):
        self.i_less = model.get_param_loc(self.p_less)
        self.i_more = model.get_param_loc(self.p_more)
        self.len = model.n_params
        if scale is not None:
            self.scale = scale

    def fun(self, x):
        return (x[self.i_more] - x[self.i_less]) * self.scale

    def jac(self, x):
        j = np.zeros_like(x)
        j[self.i_more] = self.scale
        j[self.i_less] = -self.scale
        return j

    def get_parameters(self):
        return (self.p_less, self.p_more)

    def get_binding_note(self, x):
        return f"{self.p_less!s} ≤ {self.p_more!s}"

    def as_linear_constraints(self):
        a = np.zeros([1, self.len], dtype="float64")
        a[0, self.i_more] = self.scale
        a[0, self.i_less] = -self.scale
        return [LinearConstraint(a, 0, np.inf)]

    def as_soft_penalty(self):
        from .numbamodel import d_softplus, softplus  # here to avoid circular imports

        i_more = self.i_more
        i_less = self.i_less
        scale = self.scale

        @njit(
            [
                float32(float32[:], float32, float32),
                float64(float64[:], float64, float64),
            ]
        )
        def penalty(x, intensity, sharpness=1.0):
            return -softplus(-(x[i_more] - x[i_less]) * scale * intensity, sharpness)

        @njit(
            [
                float32[:](float32[:], float32, float32),
                float64[:](float64[:], float64, float64),
            ]
        )
        def dpenalty(x, intensity, sharpness=1.0):
            j = np.zeros_like(x)
            partial = (
                d_softplus(-(x[i_more] - x[i_less]), sharpness * scale * intensity)
                * scale
                * intensity
            )
            j[i_more] = partial
            j[i_less] = -partial
            return j

        @njit(
            [
                float32[:](float32[:], float32),
                float64[:](float64[:], float64),
            ]
        )
        def dpenalty_bind(x, intensity):
            j = np.zeros_like(x)
            diff = x[i_more] - x[i_less]
            if (
                (x[i_more] + x[i_less]) == 0
                or np.absolute(diff / (x[i_more] + x[i_less])) < 5e-3
                or np.absolute(diff) < 1e-6
            ):
                partial = 0.5 * scale * intensity
                j[i_more] = partial
                j[i_less] = -partial
            return j

        return penalty, dpenalty, dpenalty_bind

    def __eq__(self, other):
        if not isinstance(other, OrderingBound):
            return False
        if (self.p_less == other.p_less) and (self.p_more == other.p_more):
            return True
        return False


class FixedBound(ParametricConstraint):
    def __init__(
        self, p, minimum=None, maximum=None, model=None, scale=10, binding_tol=1e-4
    ):
        super().__init__(binding_tol=binding_tol)
        self.p = p
        self.minimum = minimum
        self.maximum = maximum
        self.scale = 1
        if model is not None:
            self.link_model(model, scale)

    def __repr__(self):
        return f"larch.FixedBound({self.p},{self.minimum},{self.maximum})"

    def link_model(self, model, scale=None):
        self.i = model.get_param_loc(self.p)
        self.len = model.n_params
        if scale is not None:
            self.scale = scale

    def _min_fun(self, x):
        if self.minimum is not None:
            return (x[self.i] - self.minimum) * self.scale
        else:
            return np.inf

    def _max_fun(self, x):
        if self.maximum is not None:
            return (self.maximum - x[self.i]) * self.scale
        else:
            return np.inf

    def fun(self, x):
        return min(
            self._min_fun(x),
            self._max_fun(x),
        )

    def jac(self, x):
        j = np.zeros_like(x)
        if self._min_fun(x) < self._max_fun(x):
            j[self.i] = self.scale
        else:
            j[self.i] = -self.scale
        return j

    def get_parameters(self):
        return (self.p,)

    def get_binding_note(self, x):
        if self._min_fun(x) < self._max_fun(x):
            return f"{self.p!s} ≥ {self.minimum}"
        else:
            return f"{self.p!s} ≤ {self.maximum}"

    def as_linear_constraints(self):
        a = np.zeros([1, self.len], dtype="float64")
        a[0, self.i] = 1
        return [
            LinearConstraint(
                a,
                self.minimum if self.minimum is not None else -np.inf,
                self.maximum if self.maximum is not None else np.inf,
            )
        ]

    def __eq__(self, other):
        if not isinstance(other, FixedBound):
            return False
        if (
            (self.p == other.p)
            and (self.minimum == other.minimum)
            and (self.maximum == other.maximum)
        ):
            return True
        return False


#######


class ParametricConstraintList(MutableSequence):
    def __init__(self, init=None):
        self._cx = list()
        self.allow_dupes = False
        if init is not None:
            for i in init:
                if isinstance(i, ParametricConstraint):
                    self._cx.append(i)
                else:
                    raise TypeError(
                        f"members of {self.__class__.__name__} must be ParametricConstraint"
                    )

    def set_instance(self, instance):
        self._instance = instance
        for i in self._cx:
            i.link_model(self._instance)

    def __fresh(self, instance):
        newself = ParametricConstraintList()
        newself._instance = instance
        setattr(instance, self.private_name, newself)
        return newself

    def __get__(self, instance, owner):
        if instance is None:
            return self
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        return newself

    def __set__(self, instance, values):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        if newself is None:
            newself = self.__fresh(instance)
        newself.__init__(values)
        try:
            newself._instance.mangle()
        except AttributeError:
            pass
        for i in newself._cx:
            i.link_model(newself._instance)

    def __delete__(self, instance):
        try:
            newself = getattr(instance, self.private_name)
        except AttributeError:
            newself = self.__fresh(instance)
        newself.__init__()
        newself._instance = instance
        try:
            newself._instance.mangle()
        except AttributeError:
            pass

    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __getitem__(self, item):
        return self._cx[item]

    def __setitem__(self, key: int, value):
        if isinstance(value, str):
            value = interpret_contraint(value)
        if not isinstance(value, ParametricConstraint):
            raise TypeError("items must be of type ParametricConstraint")
        if self.allow_dupes or not self._is_duplicate(value):
            self._cx[key] = value
            try:
                self._instance.mangle()
            except AttributeError:
                pass
            self._cx[key].link_model(self._instance)

    def __delitem__(self, key):
        del self._cx[key]
        try:
            self._instance.mangle()
        except AttributeError:
            pass

    def __len__(self):
        return len(self._cx)

    def insert(self, index, value):
        if isinstance(value, str):
            value = interpret_contraint(value)
        if not isinstance(value, ParametricConstraint):
            raise TypeError("items must be of type ParametricConstraint")
        if self.allow_dupes or not self._is_duplicate(value):
            self._cx.insert(index, value)
            try:
                self._instance.mangle()
            except AttributeError:
                pass
            self._cx[index].link_model(self._instance)

    def __repr__(self):
        return repr(self._cx)

    def _is_duplicate(self, value):
        for i in self._cx:
            if i == value:
                return True
        return False

    def rescale(self, scale=None, index=None):
        if index is None:
            for i in self._cx:
                if scale is None:
                    i.link_model(self._instance, np.exp(np.random.uniform(-3, 3)))
                else:
                    i.link_model(self._instance, scale)
        else:
            if scale is None:
                self._cx[index].link_model(
                    self._instance, np.exp(np.random.uniform(-3, 3))
                )
            else:
                self._cx[index].link_model(self._instance, scale)


def asfloat(x):
    if "/" in x:
        x_ = x.split("/")
        return float(x_[0]) / float(x_[1])
    return float(x)


def interpret_contraint(c):
    numbr = r"\s*([-+]?\d*\.?\d*[eE]?[-+]?\d*|\d+\/\d+)\s*"
    token = r"\s*([\w#:\*&^%\$!@]+)\s*"

    n_2w = f"^{numbr}(<=|<){token}(<=|<){numbr}$"
    rx = re.match(n_2w, c)
    if rx:
        return FixedBound(
            rx.group(3), minimum=asfloat(rx.group(1)), maximum=asfloat(rx.group(5))
        )

    le_n = f"^{token}(<=|<){numbr}$"
    rx = re.match(le_n, c)
    if rx:
        return FixedBound(rx.group(1), maximum=asfloat(rx.group(3)))

    ge_n = f"^{token}(>=|>){numbr}$"
    rx = re.match(ge_n, c)
    if rx:
        return FixedBound(rx.group(1), minimum=asfloat(rx.group(3)))

    n_le = f"^{numbr}(<=|<){token}$"
    rx = re.match(n_le, c)
    if rx:
        return FixedBound(rx.group(3), minimum=asfloat(rx.group(1)))

    n_ge = f"^{numbr}(>=|>){token}$"
    rx = re.match(n_ge, c)
    if rx:
        return FixedBound(rx.group(3), maximum=asfloat(rx.group(1)))

    le = f"^{token}(<=|<){token}$"
    rx = re.match(le, c)
    if rx:
        return OrderingBound(rx.group(1), rx.group(3))

    ge = f"^{token}(>=|>){token}$"
    rx = re.match(ge, c)
    if rx:
        return OrderingBound(rx.group(3), rx.group(1))

    r_le = f"^{token}/{token}(<=|<){numbr}$"
    rx = re.match(r_le, c)
    if rx:
        return RatioBound(rx.group(1), rx.group(2), max_ratio=asfloat(rx.group(4)))

    r_ge = f"^{token}/{token}(>=|>){numbr}$"
    rx = re.match(r_ge, c)
    if rx:
        return RatioBound(rx.group(1), rx.group(2), min_ratio=asfloat(rx.group(4)))

    r_2w = f"^{numbr}(<=|<){token}/{token}(<=|<){numbr}$"
    rx = re.match(r_2w, c)
    if rx:
        return RatioBound(
            rx.group(3),
            rx.group(4),
            min_ratio=asfloat(rx.group(1)),
            max_ratio=asfloat(rx.group(6)),
        )

    raise ValueError(f"cannot interpret '{c}' as a constraint")



import unicodedata
from difflib import SequenceMatcher as _SequenceMatcher
from heapq import nlargest as _nlargest
from pathlib import Path


def strip_accents(s):
    return "".join(
        c for c in unicodedata.normalize("NFD", s) if unicodedata.category(c) != "Mn"
    )


def supercasefold(s):
    return strip_accents(s).casefold()


def case_insensitive_close_matches(word, possibilities, n=3, cutoff=0.6, excpt=None):
    """Get a list of the best "good enough" matches.

    Parameters
    ----------
    word : str
            A base string for which close matches are desired
    possibilities : Collection[str]
            Word list against which to match word.
    n : int, default 3
            Maximum number of close matches to return,  must be > 0.
    cutoff : float, default 0.6
            A float in the range [0, 1].  Possibilities
            that don't score at least that similar to `word` are ignored.

    The best (no more than n) matches among the possibilities are returned
    in a list, sorted by similarity score, most similar first.

    Examples
    --------
    >>> case_insensitive_close_matches("appel", ["ape", "apple", "peach", "puppy"])
    ['apple', 'ape']
    >>> import keyword
    >>> case_insensitive_close_matches("wheel", keyword.kwlist)
    ['while']
    >>> case_insensitive_close_matches("apples", keyword.kwlist)
    []
    >>> case_insensitive_close_matches("Accept", keyword.kwlist)
    ['except']
    >>> case_insensitive_close_matches("NonLocal", keyword.kwlist)
    ['nonlocal']
    """
    if not n > 0:
        raise ValueError(f"n must be > 0: {n!r}")
    if not 0.0 <= cutoff <= 1.0:
        raise ValueError(f"cutoff must be in [0.0, 1.0]: {cutoff!r}")
    result = []
    s = _SequenceMatcher()
    s.set_seq2(supercasefold(word))
    for x in possibilities:
        x_ = supercasefold(x)
        s.set_seq1(x_)
        if (
            s.real_quick_ratio() >= cutoff
            and s.quick_ratio() >= cutoff
            and s.ratio() >= cutoff
        ):
            result.append((s.ratio(), x))

    # Move the best scorers to head of list
    result = _nlargest(n, result)
    # Strip scores for the best n matches
    ret = [x for score, x in result]
    if excpt is not None:
        did_you_mean = "'{}' not found, did you mean {}?".format(
            word, " or ".join(f"'{s}'" for s in ret)
        )
        raise excpt(did_you_mean)
    return ret


def max_len(arg, at_least=0):
    if len(arg) == 0:
        return at_least
    try:
        return max(at_least, *map(len, arg))
    except Exception:
        try:
            return max(at_least, *map(len, *arg))
        except Exception:
            print("ERR max_len")
            print("arg=", arg)
            print("type(arg)=", type(arg))
            raise


def grid_str(arg, lineprefix="", linesuffix=""):
    x = []
    try:
        n = 0
        while 1:
            x.append(max_len(tuple(i[n] for i in arg)))
            n += 1
    except IndexError:
        n -= 1
    lines = [
        lineprefix + " ".join(t[i].ljust(x[i]) for i in range(len(t))) + linesuffix
        for t in arg
    ]
    return "\n".join(lines)


def truncate_path_for_display(p, maxlen=30):
    """
    Truncate a path if it is too long to display neatly.

    Do not use the trunacted path for actual file operations, it won't work.

    Parameters
    ----------
    p : str or pathlib.Path
    maxlen : int
            Maximum number of character to return.  This is a suggestion, as
            the basename itself is never truncated.
    """
    original = Path(p)
    if len(str(original)) <= maxlen:
        return str(original)
    original_parts_len = len(original.parts)
    trimmer = 1
    z = lambda x: str(Path("⋯", *original.parts[x:]))  # unicode equivalent '\u22EF'
    while len(z(trimmer)) > maxlen:
        trimmer += 1
        if trimmer >= original_parts_len:
            return z(original_parts_len - 1)
    return z(trimmer)



from collections.abc import Mapping

import numpy
import pandas
import seaborn
from matplotlib import colors
from xmle import Elem

try:
    import pyarrow
except ImportError:
    pyarrow = None


class DataFrameViewer(pandas.DataFrame):
    def __xml__(self):
        z = Elem("div")
        table = z.put("table", {"class": "dataframe larch_dataframe"})
        thead = table.put("thead")
        for i in range(self.index.nlevels):
            thead.put("th", text=self.index.names[i])
        for c in self.columns:
            thead.put("th", text=str(c))
        tbody = table.put("tbody")
        for i in self.index:
            row = tbody.put("tr")
            if self.index.nlevels == 1:
                row.put("th", text=str(i))
            else:
                for ii in i:
                    row.put("th", text=str(ii))
            for c in self.columns:
                cell = self.loc[i, c]
                try:
                    cell_content = cell.__xml__()
                except AttributeError:
                    if isinstance(cell, float):
                        row.put("td", text=f"{cell:.6g}")
                    else:
                        row.put("td", text=str(cell))
                else:
                    row.put("td") << cell_content
        return z

    def _repr_html_(self):
        return self.__xml__().tostring()

    def __getitem__(self, item):
        result = super().__getitem__(item)
        if type(result) == pandas.DataFrame:
            result.__class__ = DataFrameViewer
        return result

    def drop(self, *args, **kwargs):
        result = super().drop(*args, **kwargs)
        if type(result) == pandas.DataFrame:
            result.__class__ = DataFrameViewer
        return result


#####################
# DataFrame Styling
#####################


def global_background_gradient(s, m, M, cmap=None, low=0, high=0):
    if cmap is None:
        cmap = seaborn.light_palette("seagreen", as_cmap=True)
    rng = M - m
    norm = colors.Normalize(m - (rng * low), M + (rng * high))
    normed = norm(s.values)
    c = [colors.rgb2hex(x) for x in cmap(normed)]
    return ["background-color: %s" % color for color in c]


def apply_global_background_gradient(
    df, override_min=None, override_max=None, cmap=None, subset=None
):
    if cmap is None:
        seagreen = seaborn.light_palette("seagreen", as_cmap=True)
        cmap = seagreen
    df = df.apply(
        global_background_gradient,
        cmap=cmap,
        m=override_min if override_min is not None else df.data.min().min(),
        M=override_max if override_max is not None else df.data.max().max(),
        subset=subset,
    )
    return df


###################


# def dataframe_from_dict_of_series(dict_of_series):
#     dict_of_numpy_arrays = {}
#     renaming = {}
#     for k in sorted(dict_of_series.keys()):
#         v = dict_of_series[k]
#         if v in df:
#             dict_of_numpy_arrays[str(k)] = df[v].to_numpy()
#         else:
#             dict_of_numpy_arrays[str(k)] = df.eval(v).to_numpy()
#         renaming[str(k)] = k
#
#     out = pa.table(dict_of_numpy_arrays).to_pandas()
#     out.rename(columns=renaming, inplace=True)
#     return out


def columnize_with_joinable_backing(
    df, name, inplace=False, dtype=None, debug=False, backing=None
):
    try:
        return columnize(df, name, inplace=inplace, dtype=dtype, debug=debug)
    except NameError:
        if not inplace:
            return columnize(
                df.join(backing), name, inplace=False, dtype=dtype, debug=debug
            )
        else:
            raise


def columnize_with_joinable_backing_2(df, name, dtype=None, debug=False, backing=None):
    source = df
    try:
        out = columnize(df, name, inplace=False, dtype=dtype, debug=debug)
    except NameError:
        if backing is not None:
            source = df.join(backing)
            backing = None
            out = columnize(source, name, inplace=False, dtype=dtype, debug=debug)
        else:
            raise
    return out, source, backing


def _get_str_fallback(d, key):
    try:
        return d[key]
    except KeyError as err:
        try:
            return d[str(key)]
        except Exception:
            raise err from None


def columnize(df, name, inplace=True, dtype=None, debug=False, backing=None):
    """Add a computed column to a DataFrame."""
    datanames = None
    from ..roles import LinearFunction  # import here to prevent circular imports

    if isinstance(name, (list, tuple)):
        datanames = [str(_) for _ in name]
        datamap = {i: i for i in datanames}
    elif isinstance(name, Mapping):
        datanames = [str(_) for _ in name.values()]
        datamap = name
    elif isinstance(name, LinearFunction):
        datanames = [str(_.data) for _ in name]
        datamap = {i: i for i in datanames}

    if datanames is not None:
        if len(datanames) == 0:
            if inplace:
                return
            else:
                return pandas.DataFrame(index=df.index)
        columns = {}
        source = df
        for dname in set(datanames):
            out, source, backing = columnize_with_joinable_backing_2(
                source, dname, dtype, backing=backing
            )
            columns[dname] = out.to_numpy()
        content = {k: _get_str_fallback(columns, v) for (k, v) in datamap.items()}
        if pyarrow:
            # For very large dataframes, this is faster than the default pandas constructor
            dict_of_numpy_arrays = {}
            renaming = {}
            for k in sorted(content.keys()):
                dict_of_numpy_arrays[str(k)] = content[k]
                renaming[str(k)] = k
            df1 = pyarrow.table(dict_of_numpy_arrays).to_pandas()
            df1.index = df.index
            df1.rename(columns=renaming, inplace=True)
        else:
            df1 = pandas.DataFrame(content, index=df.index)
        if inplace:
            df[datamap.keys()] = df1.values
            return
        else:
            return df1

    # short circuit for natural names
    try:
        return df[name]
    except KeyError:
        pass

    from tokenize import NAME, OP, STRING, tokenize, untokenize

    from .aster import asterize

    DOT = (OP, ".")
    # COLON = (OP, ":")
    COMMA = (OP, ",")
    OBRAC = (OP, "[")
    CBRAC = (OP, "]")
    OPAR = (OP, "(")
    CPAR = (OP, ")")
    EQUAL = (OP, "=")
    from io import BytesIO

    if inplace:
        recommand = [(NAME, "df"), OBRAC, (STRING, f"'{name}'"), CBRAC, EQUAL]
    else:
        recommand = []
    recommand += [
        OPAR,
    ]
    try:
        name_encode = name.encode("utf-8")
    except AttributeError:
        name_encode = str(name).encode("utf-8")
    reader = BytesIO(name_encode)
    g = tokenize(reader.readline)
    from_df = False
    for toknum, tokval, _, _, _ in g:
        if toknum == NAME:
            pass
            if tokval in df.columns:
                # replace NAME tokens
                partial = [
                    (NAME, "df"),
                    OBRAC,
                    (STRING, f"'{tokval}'"),
                    CBRAC,
                ]
                from_df = True
                recommand.extend(partial)
            # break
            elif tokval in df.index.names:
                # replace NAME tokens with # df.index.get_level_values('NAME').values
                partial = [
                    (NAME, "df"),
                    DOT,
                    (NAME, "index"),
                    DOT,
                    (NAME, "get_level_values"),
                    OPAR,
                    (STRING, f"'{tokval}'"),
                    CPAR,
                    DOT,
                    (NAME, "values"),
                ]
                from_df = True
                recommand.extend(partial)
            else:
                # no dat contains this natural name
                # put the name back in raw, maybe it works cause it's a global, more likely
                # the exception manager below will catch it.
                recommand.append((toknum, tokval))
        else:
            recommand.append((toknum, tokval))
    if not from_df and not inplace:
        recommand = (
            [(NAME, "pandas"), DOT, (NAME, "Series")]
            + recommand
            + [COMMA, (NAME, "index"), EQUAL, (NAME, "df"), DOT, (NAME, "index")]
        )
    recommand += [
        CPAR,
    ]
    if dtype is not None:
        dtype_ = str(numpy.dtype(dtype))
        recommand += [DOT, (NAME, "astype"), OPAR, (STRING, f"'{dtype_}'"), CPAR]
    try:
        ret = untokenize(recommand).decode("utf-8")
    except Exception:
        print("<recommand>")
        print(recommand)
        print("</recommand>")
        raise
    g.close()
    reader.close()
    if debug:
        print("<ret>")
        print(ret)
        print("</ret>")
    j = asterize(ret, mode="exec" if inplace else "eval")

    try:
        if inplace:
            _result = exec(j)
        else:
            _result = eval(j)
            if not isinstance(_result, pandas.Series):
                _result = pandas.Series(_result, index=df.index)
            try:
                _result.name = name
            except AttributeError:
                pass
    except Exception as exc:
        args = exc.args
        if not args:
            arg0 = ""
        else:
            arg0 = args[0]
        arg0 = arg0 + f'\nwithin parsed command: "{name!s}"'
        if "max" in name:
            arg0 = (
                arg0
                + '\n(note to get the maximum of arrays use "fmax" not "max")'.format()
            )
        if "min" in name:
            arg0 = (
                arg0
                + '\n(note to get the minimum of arrays use "fmin" not "min")'.format()
            )
        if isinstance(exc, NameError):
            badname = str(exc).split("'")[1]
            goodnames = {
                "log",
                "exp",
                "log1p",
                "absolute",
                "fabs",
                "sqrt",
                "isnan",
                "isfinite",
                "logaddexp",
                "fmin",
                "fmax",
                "nan_to_num",
                "piece",
                "normalize",
                "hard_sigmoid",
            }
            goodnames |= set(df.columns)
            if backing is not None:
                goodnames |= set(backing.columns)
            from ..util.text_manip import case_insensitive_close_matches

            did_you_mean_list = case_insensitive_close_matches(
                badname, goodnames, n=3, cutoff=0.1, excpt=None
            )
            if len(did_you_mean_list) > 0:
                arg0 = (
                    arg0
                    + "\n"
                    + "did you mean {}?".format(
                        " or ".join(f"'{s}'" for s in did_you_mean_list)
                    )
                )
        exc.args = (arg0,) + args[1:]
        raise
    return _result


def compute(df, name, **kwargs):
    kwargs["inplace"] = False
    return columnize(df, name, **kwargs)


###################
### Typical Calibration Table


def counts_and_shares(
    tours_df,
    rows="tour_mode",
    cross="distH2Work",
    bins=None,
    stack=True,
    title=None,
    blurb=None,
    rowtotals=True,
):
    if isinstance(rows, str):
        if not isinstance(
            tours_df[rows].dtype, pandas.core.dtypes.dtypes.CategoricalDtype
        ):
            raise TypeError("column for rows must be categorical")
        cats = tours_df[rows].copy()
    else:
        if not isinstance(rows.dtype, pandas.core.dtypes.dtypes.CategoricalDtype):
            raise TypeError("column for rows must be categorical")
        cats = rows.copy()

    # firstcat, lastcat = cats.cat.categories[0], cats.cat.categories[-1]

    cats = cats.cat.add_categories(["TOTAL", "% of TOT"])

    if isinstance(cross, str):
        if bins == "cat":
            pivot_cols = pandas.Series(
                pandas.Categorical(tours_df[cross]), index=tours_df.index
            )
        elif bins is not None:
            pivot_cols = pandas.cut(
                tours_df[cross],
                bins=bins,
            )
        else:
            pivot_cols = tours_df[cross]
    elif isinstance(cross, pandas.Series):
        if isinstance(cross.dtype, pandas.core.dtypes.dtypes.CategoricalDtype):
            pivot_cols = cross
        else:
            raise TypeError("if a Series, cross must be a Categorical Series")
    else:
        raise TypeError(f"bad cross of type {type(cross)}")

    result1 = (
        pandas.pivot_table(
            tours_df,
            values="personId",
            index=cats,
            columns=pivot_cols,
            aggfunc=len,
        )
        .fillna(0)
        .astype(int)
    )

    result1.columns = result1.columns.astype(str)

    if rowtotals:
        result1.loc[:, "TOTAL"] = result1.sum(1)

    result2 = result1 / result1.sum(0)
    result2.loc["TOTAL", :] = numpy.nan
    result2.loc["% of TOT", :] = numpy.nan

    if rowtotals:
        result1.loc["TOTAL", :] = result1.sum(0)
        result1.loc["% of TOT", :] = (
            -result1.loc["TOTAL", :] * 2 / result1.loc["TOTAL", :].sum()
        )
    else:
        result1.loc["TOTAL", :] = result1.sum(0)
        result1.loc["% of TOT", :] = (
            -result1.loc["TOTAL", :] / result1.loc["TOTAL", :].sum()
        )

    def pct_or_thousands(x):
        if x < 0.0:
            return f"{-x:.1%}"
        else:
            return f"{x:,.0f}"

    def hide_nan_else_pct(x):
        if pandas.isnull(x):
            return ""
        else:
            return f"{x:.2%}"

    if stack:
        result = pandas.concat(
            [result1, result2],
            axis=1,
            keys=["Counts", "Shares"],
            sort=False,
        )
        result = result.style.format(
            {
                **{i: pct_or_thousands for i in result.columns if i[0] == "Counts"},
                **{i: hide_nan_else_pct for i in result.columns if i[0] == "Shares"},
            }
        )

    else:
        result = (
            result1.style.format("{:,.0f}"),
            result2.style.format("{:.2%}"),
        )

    # try:
    # 	result = apply_global_background_gradient(
    # 		result,
    # 		override_max=0.75,
    # 		override_min=0.01,
    # 		# subset=[i for i in result.columns if i[0]=='Shares'],
    # 		subset=pandas.IndexSlice[firstcat:lastcat, [i for i in result.columns if i[0] == 'Shares']]
    # 	)
    # except ValueError:
    # 	pass

    result.map(
        lambda y: "text-align:right;",
        # subset=pandas.IndexSlice[firstcat:lastcat, [i for i in result.columns if i[0]=='Counts']]
    )

    if title is None:
        return result
    else:
        from IPython.display import HTML, display, display_html

        display_html(HTML(f"<h2>{title}</h2>"))
        if blurb is not None:
            display_html(HTML(f"<p>{blurb}</p>"))
        display(result)



def example(extract='m', estimate=False):
    import larch as lx

    lx.__version__

    m = lx.example(17)
    m.compute_engine = "numba"

    motorized = m.graph.new_node(
        parameter="mu_motor", children=[1, 2, 3, 4], name="Motorized"
    )
    nonmotorized = m.graph.new_node(
        parameter="mu_nonmotor", children=[5, 6], name="Nonmotorized"
    )

    m.ordering = (
        (
            "CostbyInc",
            "costbyincome",
        ),
        (
            "TravelTime",
            ".*time.*",
            ".*dist.*",
        ),
        (
            "Household",
            "hhinc.*",
            "vehbywrk.*",
        ),
        (
            "Zonal",
            "wkcbd.*",
            "wkempden.*",
        ),
        (
            "ASCs",
            "ASC.*",
        ),
    )

    m.graph

    mj = m.copy()
    mj.compute_engine = "jax"
    if not estimate:
        if isinstance(extract, str):
            return locals()[extract]
        else:
            _locals = locals()
            return [_locals.get(i) for i in extract]
    result = m.maximize_loglike(method="bhhh")

    m.calculate_parameter_covariance()
    m.parameter_summary()

    resultj = mj.maximize_loglike(stderr=True)
    if isinstance(extract, str):
        return locals()[extract]
    else:
        _locals = locals()
        return [_locals.get(i) for i in extract]



import operator
from collections.abc import Mapping

import numpy
import pandas
from sklearn.preprocessing import OneHotEncoder


def one_hot_encode(vector, dtype="float32", categories=None, index=None):
    if isinstance(vector, pandas.Series):
        index = vector.index
        vector = vector.values

    encoder = OneHotEncoder(
        categories="auto"
        if categories is None
        else [
            categories,
        ],
        sparse=False,
        dtype=dtype,
    ).fit(vector.reshape(-1, 1))

    return pandas.DataFrame(
        data=encoder.transform(vector.reshape(-1, 1)),
        columns=encoder.categories_,
        index=index,
    )


def periodize(
    values,
    mapping=None,
    default=None,
    right=True,
    left=True,
    **kwargs,
):
    """
    Label sections of a continuous variable.

    This function contrasts with `pandas.cut` in that
    there can be multiple non-contiguous sections of the
    underlying continuous interval that obtain the same
    categorical value.

    Parameters
    ----------
    values : array-like
            The values to label. If given as a pandas.Series,
            the returned values will also be a Series,
            with a categorical dtype.
    mapping : Collection or Mapping
            A mapping, or a collection of 2-tuples giving
            key-value pairs (not necessarily unique keys).
            The keys (or first values) will be the new values,
            and the values (or second values) are 2-tuples
            giving upper and lower bounds.
    default : any, default None
            Keys not inside any labeled interval will get
            this value instead.
    right : bool, default True
            Whether to include the upper bound[s] in the
            intervals for labeling.
    left : bool, default True
            Whether to include the lower bound[s] in the
            intervals for labeling.
    **kwargs :
            Are added to `mapping`.

    Returns
    -------
    array-like

    Example
    -------
    >>> import pandas
    >>> h = pandas.Series(range(1,24))
    >>> periodize(h, default='OP', AM=(6.5, 9), PM=(16, 19))
    0     OP
    1     OP
    2     OP
    3     OP
    4     OP
    5     OP
    6     AM
    7     AM
    8     AM
    9     OP
    10    OP
    11    OP
    12    OP
    13    OP
    14    OP
    15    PM
    16    PM
    17    PM
    18    PM
    19    OP
    20    OP
    21    OP
    22    OP
    dtype: category
    Categories (3, object): ['AM', 'OP', 'PM']
    """
    if mapping is None:
        mapping = []

    if isinstance(mapping, Mapping):
        mapping = list(mapping.items())

    mapping.extend(kwargs.items())

    if isinstance(values, pandas.Series):
        x = pandas.Series(index=values.index, data=default)
    else:
        x = numpy.full(values.shape, default)

    if right:
        rop = operator.le
    else:
        rop = operator.lt

    if left:
        lop = operator.ge
    else:
        lop = operator.gt

    for k, (lowerbound, upperbound) in mapping:
        if lowerbound is None:
            lowerbound = -numpy.inf
        if upperbound is None:
            upperbound = numpy.inf
        x[lop(values, lowerbound) & rop(values, upperbound)] = k

    if isinstance(x, pandas.Series):
        x = x.astype("category")

    return x



try:
    from PIL import Image, ImageChops, ImageDraw
    from selenium import webdriver
except ImportError:
    pass  # Nothing in this module will work

import os
from io import BytesIO


class _ScreenCapper:
    def __init__(self):
        options = webdriver.ChromeOptions()
        options.add_argument("headless")
        options.add_argument("window-size=1024x768")
        options.add_argument("disable-overlay-scrollbar")
        self.driver = webdriver.Chrome(chrome_options=options)

    def __del__(self):
        self.driver.close()

    def _screenshot_formatter(self, url, window_size=(1024, 768), full_height=False):
        self.driver.get(url)
        if window_size[1] is None or full_height:
            # Full height window
            self.driver.maximize_window()
            height = self.driver.execute_script("return document.body.scrollHeight")
            window_size = window_size[0], height
            self.driver.set_window_size(*window_size)

    def screenshot_to_file(
        self,
        url,
        file_name,
        log=None,
        output_dir="./",
        window_size=(1024, 768),
        full_height=False,
    ):
        if log is not None:
            log("Capturing", url, "screenshot as", file_name, "...")
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        self._screenshot_formatter(url, window_size, full_height=full_height)
        self.driver.save_screenshot(file_name)
        if log is not None:
            log("Done capturing", url)

    def screenshot_to_base64(
        self, url, log=None, window_size=(1024, 768), full_height=False
    ):
        if log is not None:
            log("Capturing", url, "screenshot as base64...")
        self._screenshot_formatter(url, window_size, full_height=full_height)
        result = self.driver.get_screenshot_as_base64()
        if log is not None:
            log("Done capturing", url)
        return result


_grabber = None


def screenshot(
    url,
    file_name=None,
    log=None,
    output_dir="./",
    window_size=(1024, 768),
    full_height=False,
):
    global _grabber
    if _grabber is None:
        _grabber = _ScreenCapper()
    if file_name is None:
        return _grabber.screenshot_to_base64(
            url, log=log, window_size=window_size, full_height=full_height
        )
    else:
        _grabber.screenshot_to_file(
            url,
            file_name,
            log=log,
            output_dir=output_dir,
            window_size=window_size,
            full_height=full_height,
        )


# https://gist.github.com/rinchik/d023578a705d6d4a5b12e235c5a9df9a#file-averageregionbrightness-py


def _process_region(image, x, y, width, height, tolerance=100):
    region_total = 0

    # This can be used as the sensitivity factor, the larger it is the less sensitive the comparison
    factor = tolerance

    for coordinateY in range(y, y + height):
        for coordinateX in range(x, x + width):
            try:
                pixel = image.getpixel((coordinateX, coordinateY))
                region_total += sum(pixel) / 4
            except Exception:
                return

    return region_total / factor


def diff_images(imagefile1, imagefile2, return_blocked1=False, tolerance=100):
    """
    Compare two images and return the number of mismatched blocks.

    Parameters
    ----------
    imagefile1, imagefile2 : images
            Two images to compare.  They should be the same size.  Give by
            filename or with raw data.
    return_blocked1 : bool, default False
            If true, also return an image `Elem` based on imagefile1 showing
            where the mismatch blocks are.
    tolerance : int
            This can be used as the sensitivity factor, the larger it is the
            less sensitive the comparison

    Returns
    -------
    int
            The number of mismatched blocks
    Elem
            optional, see return_blocked1
    """
    screenshot_staging = Image.open(imagefile1)
    screenshot_production = Image.open(imagefile2)
    columns = 60
    rows = 80
    screen_width, screen_height = screenshot_staging.size

    block_width = ((screen_width - 1) // columns) + 1  # this is just a division ceiling
    block_height = ((screen_height - 1) // rows) + 1

    mismatch_blocks = 0

    if return_blocked1:
        degenerate = Image.new(
            screenshot_staging.mode, screenshot_staging.size, "#FFFFFF"
        )

        if "A" in screenshot_staging.getbands():
            degenerate.putalpha(screenshot_staging.getchannel("A"))

        screenshot_staging_1 = Image.blend(degenerate, screenshot_staging, 0.25)

    for y in range(0, screen_height, block_height + 1):
        for x in range(0, screen_width, block_width + 1):
            region_staging = _process_region(
                screenshot_staging, x, y, block_width, block_height, tolerance=tolerance
            )
            region_production = _process_region(
                screenshot_production,
                x,
                y,
                block_width,
                block_height,
                tolerance=tolerance,
            )

            if (
                region_staging is not None
                and region_production is not None
                and region_production != region_staging
            ):
                mismatch_blocks += 1
                if return_blocked1:
                    draw = ImageDraw.Draw(screenshot_staging_1)
                    draw.rectangle(
                        (x, y, x + block_width, y + block_height), outline="red"
                    )

    if return_blocked1:
        buffered = BytesIO()
        screenshot_staging_1.save(buffered, format="PNG")
        from xmle import Elem

        blocked1 = Elem.from_png_raw(buffered.getvalue())
        return mismatch_blocks, blocked1

    return mismatch_blocks


# https://stackoverflow.com/questions/10615901/trim-whitespace-using-pil/10616717


def trim_uniform_border(im):
    """
    Trim a border from an image.

    Parameters
    ----------
    im : PIL.Image.Image

    Returns
    -------
    PIL.Image.Image

    """
    bg = Image.new(im.mode, im.size, im.getpixel((0, 0)))
    diff = ImageChops.difference(im, bg)
    diff = ImageChops.add(diff, diff, 2.0, -100)
    bbox = diff.getbbox()
    if bbox:
        return im.crop(bbox)



import numpy as np
import xarray as xr
from xarray import Dataset

_GROUPID = "_groupid_"
_INGROUP = "_ingroup_"


def random_groups():
    pass


def _group_breaks(groupid):
    group_breaks = np.append(
        np.insert(np.where(np.diff(groupid))[0] + 1, 0, 0), len(groupid)
    )
    lens = np.diff(group_breaks)
    pad_len = np.max(lens) - lens
    where_to_pad = np.repeat(group_breaks[1:], pad_len)
    ids = groupid[group_breaks[:-1]]
    return lens, where_to_pad, ids


def fold(
    arr,
    groupid,
    fill_value=0,
    group_dim="groupid",
    ingroup_dim="ingroup",
    lens=None,
    where_to_pad=None,
    gids=None,
):
    try:
        arr_GROUPID = arr.attrs.get(_GROUPID, None)
    except AttributeError:
        pass
    else:
        if arr_GROUPID == group_dim:
            # assume already folded
            # TODO check
            return arr
    if arr.shape[0] < groupid.size:
        # assume already folded
        # TODO check
        return arr
    if lens is None or where_to_pad is None:
        lens, where_to_pad, gids = _group_breaks(groupid)
        # lens = np.diff(group_breaks)
        # pad_len = np.max(lens) - lens
        # where_to_pad = np.repeat(group_breaks[1:], pad_len)
    padding_value = np.full(
        arr.shape[1:],
        fill_value,
    )
    reshape_to = [-1, np.max(lens), *arr.shape[1:]]
    bigger = np.insert(np.asarray(arr), where_to_pad, padding_value, axis=0)
    try:
        padded_arr = bigger.reshape(reshape_to)
    except ValueError:
        if bigger.size == 0:
            padded_arr = bigger.reshape(
                [int(bigger.shape[0] / np.max(lens)), np.max(lens), *arr.shape[1:]]
            )
        else:
            raise
    if isinstance(arr, xr.DataArray):
        return arr.__class__(
            padded_arr,
            dims=[group_dim, ingroup_dim, *arr.dims[1:]],
            coords={group_dim: np.asarray(gids)} if gids is not None else {},
            name=arr.name,
        )
    return padded_arr


def fold_dataset(
    d,
    groupid="groupid",
    group_dim="groupid",
    ingroup_dim="ingroup",
    vars=("ca", "co", "ch", "av", "caseid"),
    collapse_zero_variance=False,
):
    try:
        if group_dim in d.dims and ingroup_dim in d.dims:
            return d  # already folded!
    except AttributeError:
        raise
    if isinstance(groupid, str):
        if groupid not in d:
            return d
        g = d[groupid]
    else:
        g = groupid
    out = Dataset()
    lens, where_to_pad, gids = _group_breaks(g)
    if vars is None:
        vars = [i for i in d.variables if i not in (group_dim, ingroup_dim)]
    for i in vars:
        if i in d:
            temp = fold(
                d[i],
                g,
                group_dim=group_dim,
                ingroup_dim=ingroup_dim,
                lens=lens,
                where_to_pad=where_to_pad,
                gids=gids,
            )
            out = out.assign({i: temp})
    for c in d.coords:
        if d.dc.CASEID not in d[c].dims:
            out.coords[c] = d.coords[c]
    out.dc.ALTID = d.dc.ALTID
    try:
        out.dc.GROUPID = group_dim
        out.dc.INGROUP = ingroup_dim
    except Exception:
        pass
    return out


# def folder(cls, ds, fold_on, crack=True, avail="_avail_", fill_missing=None):
#     """
#     Construct a Dataset from an idca-format DataFrame.
#
#     This method loads the data as dense arrays.
#
#     Parameters
#     ----------
#     ds : Dataset
#         The input data should be an idca-format or idce-format DataFrame,
#         with the caseid's and altid's in a two-level pandas MultiIndex.
#     crack : bool, default True
#         If True, the `dissolve_zero_variance` method is applied before
#         repairing dtypes, to ensure that missing value are handled
#         properly.
#     avail : str, default '_avail_'
#         When the imported data is sparse then
#         an availability indicator is computed and given this name.
#     fill_missing : scalar or Mapping, optional
#         Fill values to use for missing values when imported data is
#         sparse.  Give a single value to use
#         globally, or a mapping of {variable: value} or {dtype: value}.
#
#     Returns
#     -------
#     Dataset
#
#     """
#     # if df.index.nlevels != 2:
#     #     raise ValueError("source idca dataframe must have a two "
#     #                      "level MultiIndex giving case and alt id's")
#     # caseidname, altidname = df.index.names
#
#     # check altids are integers, if they are not then fix it
#     # if df.index.levels[1].dtype.kind != 'i':
#     #     if altnames is None:
#     #         altnames = df.index.levels[1]
#     #         df.index = df.index.set_levels(np.arange(1, len(altnames) + 1), level=1)
#     #     else:
#     #         new_index = df.index.get_level_values(1).astype(pd.CategoricalDtype(altnames))
#     #         df.index = df.index.set_codes(
#     #             new_index.codes, level=1
#     #         ).set_levels(np.arange(1, len(altnames) + 1), level=1)
#
#     ds = cls()(df, caseid=caseidname, alts=altidname)
#     if crack:
#         ds = ds.dc.dissolve_zero_variance()
#     ds = ds.set_dtypes(df)
#     if altnames is not None:
#         ds = ds.dc.set_altnames(altnames)
#     if avail not in ds and len(df) < ds.dc.n_cases * ds.dc.n_alts:
#         av = (
#             xr.DataArray.from_series(pd.Series(1, index=df.index))
#             .fillna(0)
#             .astype(np.int8)
#         )
#         ds[avail] = av
#         if fill_missing is not None:
#             if isinstance(fill_missing, Mapping):
#                 for k, i in ds.items():
#                     if ds.dc.ALTID not in i.dims:
#                         continue
#                     if k not in fill_missing and i.dtype not in fill_missing:
#                         continue
#                     filler = fill_missing.get(k, fill_missing[i.dtype])
#                     ds[k] = i.where(ds["_avail_"] != 0, filler)
#             else:
#                 for k, i in ds.items():
#                     if ds.dc.ALTID not in i.dims:
#                         continue
#                     ds[k] = i.where(ds["_avail_"] != 0, fill_missing)
#     return ds


def dissolve_zero_variance(self, dim, mask=None, inplace=False):
    """
    Dissolve dimension on variables where it has no variance.

    This method is convenient to convert variables that have
    been loaded as |idca| or |idce| format into |idco| format where
    appropriate.

    Parameters
    ----------
    dim : str, optional
        The name of the dimension to potentially dissolve.
    inplace : bool, default False
        Whether to dissolve variables in-place.

    Returns
    -------
    Dataset
    """
    if inplace:
        obj = self
    else:
        obj = self.copy()
    for k in obj.variables:
        if obj[k].dtype.kind in {"U", "S", "O"}:
            continue
        if dim in obj[k].dims:
            try:
                if mask:
                    dissolve = obj[k].where(self[mask]).std(dim=dim).max() < 1e-10
                else:
                    dissolve = obj[k].std(dim=dim).max() < 1e-10
            except TypeError:
                pass
            else:
                if dissolve:
                    if mask:
                        obj[k] = (
                            obj[k].where(self[mask]).min(dim=dim).astype(obj[k].dtype)
                        )
                    else:
                        obj[k] = obj[k].min(dim=dim)
    return obj



import glob
import json
import os
import textwrap


def rip():
    targets = [
        "docs/examples/mtc/*.ipynb",
        "docs/examples/swissmetro/*.ipynb",
        "docs/examples/exampville/*.ipynb",
        "docs/examples/xlogit-artificial/*.ipynb",
    ]

    targets = [
        os.path.normpath(os.path.join(os.path.dirname(__file__), "..", t))
        for t in targets
    ]

    generator_dest = os.path.normpath(
        os.path.join(
            os.path.dirname(__file__),
            "..",
            "src",
            "larch",
            "examples",
            "generated",
        )
    )
    os.makedirs(generator_dest, exist_ok=True)
    print(f"generator_dest={generator_dest}")

    with open(os.path.join(generator_dest, ".gitignore"), "w", encoding="utf8") as f:
        f.write(".gitignore\n*.py")

    import_stmts = []
    assign_stmts = []

    for target in targets:
        for examplefile in glob.glob(target):
            print(f"examplefile={examplefile}")
            example = os.path.basename(examplefile).replace(".ipynb", "")
            example = example.replace("-", "_")
            print(f"  example={example}")
            with open(
                os.path.join(generator_dest, f"_{example}.py"), "w", encoding="utf8"
            ) as f:
                f.write("def example(extract='m', estimate=False):\n")
                with open(examplefile, encoding="utf8") as s:
                    sourcefilecontent = s.read()
                y = json.loads(sourcefilecontent)
                source1 = []
                source2 = []
                trip = False
                for i in y["cells"]:
                    if (
                        i["cell_type"] == "code"
                        and not i["metadata"].get("doc_only", False)
                        and not i["metadata"].get("remove-cell", False)
                        and "remove-cell" not in i.get("metadata", {}).get("tags", [])
                    ):
                        s = "".join(i["source"])
                        if ".maximize_loglike(" in s:
                            trip = True
                        if trip:
                            source2.append(s)
                        else:
                            source1.append(s)
                sourcecode1 = "\n\n".join(source1)
                sourcecode2 = "\n\n".join(source2)
                f.write(textwrap.indent(sourcecode1, "    "))
                mid = """
                if not estimate:
                    if isinstance(extract, str):
                        return locals()[extract]
                    else:
                        _locals = locals()
                        return [_locals.get(i) for i in extract]
                """
                f.write(textwrap.indent(textwrap.dedent(mid), "    "))
                f.write(textwrap.indent(sourcecode2, "    "))
                post = """
                if isinstance(extract, str):
                    return locals()[extract]
                else:
                    _locals = locals()
                    return [_locals.get(i) for i in extract]
                """
                f.write(textwrap.indent(textwrap.dedent(post), "    "))

            ex_n = example.split("_")[0]
            try:
                ex_n = int(ex_n)
            except ValueError:
                pass
            import_stmts.append(f"from ._{example} import example as _{example}")
            assign_stmts.append(f"ex[{ex_n!r}] = _{example}")

    with open(os.path.join(generator_dest, "__init__.py"), "w", encoding="utf8") as gen:
        gen.write(
            textwrap.dedent(
                """
        # Example code extracted automatically from documentation.
        # Do not edit these files manually.
        """
            )
        )
        for s in import_stmts:
            gen.write(s)
            gen.write("\n")
        gen.write("\nex = {}\n")
        for s in assign_stmts:
            gen.write(s)
            gen.write("\n")


if __name__ == "__main__":
    rip()



#
# Adapted from the base pandas Styler.background_gradient
# https://github.com/pandas-dev/pandas/blob/v1.0.2/pandas/io/formats/style.py
#
# BSD 3-Clause License
#
# Copyright (c) 2008-2012, AQR Capital Management, LLC, Lambda Foundry, Inc. and PyData Development Team
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of the copyright holder nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
#


import numpy as np
import pandas as pd
from pandas.api.types import is_list_like
from pandas.core.dtypes.generic import ABCSeries
from pandas.core.indexes.api import Index
from pandas.io.formats.style import Styler, _mpl


# the public IndexSlicerMaker
class _IndexSlice:
    """
    Create an object to more easily perform multi-index slicing.

    See Also
    --------
    MultiIndex.remove_unused_levels : New MultiIndex with no unused levels.

    Notes
    -----
    See :ref:`Defined Levels <advanced.shown_levels>`
    for further info on slicing a MultiIndex.

    Examples
    --------
    >>> midx = pd.MultiIndex.from_product([['A0','A1'], ['B0','B1','B2','B3']])
    >>> columns = ['foo', 'bar']
    >>> dfmi = pd.DataFrame(np.arange(16).reshape((len(midx), len(columns))),
    ...                     index=midx, columns=columns)

    Using the default slice command:

    >>> dfmi.loc[(slice(None), slice('B0', 'B1')), :]
               foo  bar
    A0 B0    0    1
       B1    2    3
    A1 B0    8    9
       B1   10   11

    Using the IndexSlice class for a more intuitive command:

    >>> idx = pd.IndexSlice
    >>> dfmi.loc[idx[:, 'B0':'B1'], :]
               foo  bar
    A0 B0    0    1
       B1    2    3
    A1 B0    8    9
       B1   10   11
    """

    def __getitem__(self, arg):
        return arg


IndexSlice = _IndexSlice()


def _non_reducing_slice(slice_):
    """
    Ensure that a slice doesn't reduce to a Series or Scalar.

    Any user-passed `subset` should have this called on it
    to make sure we're always working with DataFrames.
    """
    # default to column slice, like DataFrame
    # ['A', 'B'] -> IndexSlices[:, ['A', 'B']]
    kinds = (ABCSeries, np.ndarray, Index, list, str)
    if isinstance(slice_, kinds):
        slice_ = IndexSlice[:, slice_]

    def pred(part) -> bool:
        """
        Check if slice does not reduce.

        Returns
        -------
        bool
                True if slice does *not* reduce,
                False if `part` is a tuple.
        """
        # true when slice does *not* reduce, False when part is a tuple,
        # i.e. MultiIndex slice
        return (isinstance(part, slice) or is_list_like(part)) and not isinstance(
            part, tuple
        )

    if not is_list_like(slice_):
        if not isinstance(slice_, slice):
            # a 1-d slice, like df.loc[1]
            slice_ = [[slice_]]
        else:
            # slice(a, b, c)
            slice_ = [slice_]  # to tuplize later
    else:
        slice_ = [part if pred(part) else [part] for part in slice_]
    return tuple(slice_)


def _maybe_numeric_slice(df, slice_, include_bool=False):
    """
    Check for numeric dtypes in the DataFrame.

    Want nice defaults for background_gradient that don't break
    with non-numeric data. But if slice_ is passed go with that.
    """
    if slice_ is None:
        dtypes = [np.number]
        if include_bool:
            dtypes.append(bool)
        slice_ = IndexSlice[:, df.select_dtypes(include=dtypes).columns]
    return slice_


class OverspecView(Styler):
    def __init__(self, data, *args, cmap="YlOrRd", high=0.3, **kwargs):
        if isinstance(data, (tuple, list)):
            data_ = pd.concat(
                [
                    pd.DataFrame(
                        eigvec, index=eigpar, columns=[f"({n + 1}) {eigval:.4g}"]
                    )
                    for n, (eigval, eigpar, eigvec) in enumerate(data)
                ],
                axis=1,
            )
        elif data is None:
            data_ = pd.DataFrame()
        else:
            data_ = data
        super().__init__(data_, *args, **kwargs)
        self.absolute_background_gradient(cmap=cmap, high=high)

    def absolute_background_gradient(
        self,
        cmap="YlOrRd",
        low=0,
        high=0,
        axis=0,
        subset=None,
        text_color_threshold=0.408,
        vmin: float | None = 0,
        vmax: float | None = None,
    ):
        """
        Color the background in a gradient style.

        Color the background in a gradient style,
        based on the magnitude (absolute value).
        The background color is determined according
        to the data in each column (optionally row).
        Requires matplotlib.


        Parameters
        ----------
        cmap : str or colormap
                Matplotlib colormap.
        low : float
                Compress the range by the low.
        high : float
                Compress the range by the high.
        axis : {0 or 'index', 1 or 'columns', None}, default 0
                Apply to each column (``axis=0`` or ``'index'``), to each row
                (``axis=1`` or ``'columns'``), or to the entire DataFrame at once
                with ``axis=None``.
        subset : IndexSlice
                A valid slice for ``data`` to limit the style application to.
        text_color_threshold : float or int
                Luminance threshold for determining text color. Facilitates text
                visibility across varying background colors. From 0 to 1.
                0 = all text is dark colored, 1 = all text is light colored.
        vmin : float, default 0
                Minimum data value that corresponds to colormap minimum value.
                When None, the minimum value of the data will be used.
        vmax : float, optional
                Maximum data value that corresponds to colormap maximum value.
                When None (default): the maximum value of the data will be used.

        Returns
        -------
        self : Styler

        Raises
        ------
        ValueError
                If ``text_color_threshold`` is not a value from 0 to 1.

        Notes
        -----
        Set ``text_color_threshold`` or tune ``low`` and ``high`` to keep the
        text legible by not using the entire range of the color map. The range
        of the data is extended by ``low * (x.max() - x.min())`` and ``high *
        (x.max() - x.min())`` before normalizing.
        """
        subset = _maybe_numeric_slice(self.data, subset)
        subset = _non_reducing_slice(subset)
        self.apply(
            self._absolute_background_gradient,
            cmap=cmap,
            subset=subset,
            axis=axis,
            low=low,
            high=high,
            text_color_threshold=text_color_threshold,
            vmin=vmin,
            vmax=vmax,
        )
        return self

    @staticmethod
    def _absolute_background_gradient(
        s,
        cmap="YlOrRd",
        low=0,
        high=0,
        text_color_threshold=0.408,
        vmin: float | None = 0,
        vmax: float | None = None,
    ):
        """Color background in a range according to the data."""
        if (
            not isinstance(text_color_threshold, (float, int))
            or not 0 <= text_color_threshold <= 1
        ):
            msg = "`text_color_threshold` must be a value from 0 to 1."
            raise ValueError(msg)

        with _mpl(Styler.background_gradient) as (plt, colors):
            smin = np.nanmin(np.absolute(s.to_numpy())) if vmin is None else vmin
            smax = np.nanmax(np.absolute(s.to_numpy())) if vmax is None else vmax
            rng = smax - smin
            # extend lower / upper bounds, compresses color range
            norm = colors.Normalize(smin - (rng * low), smax + (rng * high))
            # matplotlib colors.Normalize modifies inplace?
            # https://github.com/matplotlib/matplotlib/issues/5427
            rgbas = plt.cm.get_cmap(cmap)(norm(np.absolute(s.to_numpy(dtype=float))))
            rgbas[np.isnan(s), :3] = 1.0

            def relative_luminance(rgba):
                """
                Calculate relative luminance of a color.

                The calculation adheres to the W3C standards
                (https://www.w3.org/WAI/GL/wiki/Relative_luminance)

                Parameters
                ----------
                color : rgb or rgba tuple

                Returns
                -------
                float
                        The relative luminance as a value from 0 to 1
                """
                r, g, b = (
                    x / 12.92 if x <= 0.03928 else ((x + 0.055) / 1.055**2.4)
                    for x in rgba[:3]
                )
                return 0.2126 * r + 0.7152 * g + 0.0722 * b

            def css(rgba):
                dark = relative_luminance(rgba) < text_color_threshold
                text_color = "#f1f1f1" if dark else "#000000"
                return f"background-color: {colors.rgb2hex(rgba)};color: {text_color};"

            if s.ndim == 1:
                return [css(rgba) for rgba in rgbas]
            else:
                return pd.DataFrame(
                    [[css(rgba) for rgba in row] for row in rgbas],
                    index=s.index,
                    columns=s.columns,
                )



class SingleParameter:
    def __set_name__(self, owner, name):
        self.name = name
        self.private_name = "_" + name

    def __get__(self, instance, instancetype=None):
        """
        Get attribute from the instance.

        Parameters
        ----------
        instance : Any
            Instance of parent class that has `self` as a member.
        instancetype : class
            Class of `instance`.
        """
        if instance is None:
            return self
        return getattr(instance, self.private_name, None)

    def __set__(self, instance, value):
        """
        Get attribute on the instance.

        Parameters
        ----------
        instance : Any
            Instance of parent class that has `self` as a member.
        value : Any
            Value being set.
        """
        existing_value = self.__get__(instance)
        if value is not None:
            value = str(value)
        if existing_value != value:
            setattr(instance, self.private_name, value)
            try:
                instance.mangle()
            except AttributeError:
                pass

    def __delete__(self, instance):
        self.__set__(instance, None)



import numpy
import numpy as np
import pandas
import pandas as pd
from xarray import Dataset

from .._optional import pyplot as plt
from ..dataset import DataTree
from ..model.basemodel import BaseModel
from .plotting import plot_as_svg_xhtml

colors = [
    "#1F77B4",
    "#FF7F0E",
    "#2CA02C",
    "#D62728",
    "#9467BD",
    "#8C564B",
    "#CFECF9",
    "#7F7F7F",
    "#BCBD22",
    "#17BECF",
]


def pseudo_bar_data(x_bins, y, gap=0):
    """
    Generate pseudo bar data.

    Parameters
    ----------
    x_bins : array-like, shape=(N+1,)
            The bin boundaries
    y : array-like, shape=(N,)
            The bar heights

    Returns
    -------
    x, y
    """
    # midpoints = (x_bins[1:] + x_bins[:-1]) / 2
    # widths = x_bins[1:] - x_bins[:-1]
    if gap:
        x_doubled = numpy.zeros(((x_bins.shape[0] - 1) * 4), dtype=numpy.float)
        x_doubled[::4] = x_bins[:-1]
        x_doubled[1::4] = x_bins[:-1]
        x_doubled[2::4] = x_bins[1:] - gap
        x_doubled[3::4] = x_bins[1:] - gap
        y_doubled = numpy.zeros(((y.shape[0]) * 4), dtype=y.dtype)
        y_doubled[1::4] = y
        y_doubled[2::4] = y
    else:
        x_doubled = numpy.zeros((x_bins.shape[0] - 1) * 2, dtype=x_bins.dtype)
        x_doubled[::2] = x_bins[:-1]
        x_doubled[1::2] = x_bins[1:]
        y_doubled = numpy.zeros((y.shape[0]) * 2, dtype=y.dtype)
        y_doubled[::2] = y
        y_doubled[1::2] = y
    return x_doubled, y_doubled


def distribution_figure(
    x,
    probability,
    choices=None,
    availability=None,
    xlabel=None,
    ylabel="Relative Frequency",
    style="hist",
    bins=None,
    pct_bins=20,
    range=None,
    prob_label="Modeled",
    obs_label="Observed",
    bw_method=None,
    discrete=None,
    ax=None,
    format="ax",
    accumulator=False,
    xscale=None,
    xmajorticks=None,
    xminorticks=None,
    coincidence_ratio=False,
    **kwargs,
):
    """
    Generate a figure of observed and modeled choices over a range of variable values.

    Parameters
    ----------
    x : array-like
            An array giving values for some variable.
    probability : array-like
            The pre-calculated probability array for all cases in this analysis.
            Must be the same shape as `x`.
    choices : array-like, optional
            The observed choices array for all cases in this analysis. If provided,
            must be the same shape as `x`.
    availability : array-like, optional
            The availability array for all cases in this analysis. If provided,
            must be the same shape as `x`.
    xlabel : str, optional
            A label to use for the x-axis of the resulting figure.  If not given,
            the value of `x.name` is used if it is defined.  Set to `False` to omit the
            x-axis label even if `x.name` is defined.
    ylabel : str, default "Relative Frequency"
            A label to use for the y-axis of the resulting figure.
    style : {'hist', 'kde'}
            The style of figure to produce, either a histogram or a kernel density plot.
    bins : int, default 25
            The number of bins to use, only applicable to histogram style.
    range : 2-tuple, optional
            A range to truncate the figure.
    prob_label : str, default "Modeled"
            A label to put in the legend for the modeled probabilities
    obs_label : str, default "Observed"
            A label to put in the legend for the observed choices
    subselector : str or array-like, optional
            A filter to apply to cases. If given as a string, this is loaded from the
            model's `dataservice` as an `idco` variable.
    ax : matplotlib.Axes, optional
            If given, the figure will be drawn on these axes and they will be returned,
            otherwise new blank axes are used to draw the figure.
    format : {'ax', 'figure', 'svg'}, default 'figure'
            How to return the result if it is a figure. The default is to return
            the raw matplotlib Axes instance.  Change this to `svg` to get a SVG
            rendering as an xmle.Elem.
    accumulator : bool, default False
            Add an net cumulative trend on the bottom.

    Returns
    -------
    Elem or Axes
            Returns `ax` if given as an argument, otherwise returns a rendering as an Elem
    """
    _coincidence_ratio = None

    if xlabel is None:
        try:
            xlabel = x.name
        except AttributeError:
            pass

    discrete_values = None
    if discrete:
        discrete_values = numpy.unique(x)
    elif discrete is None:
        from .histograms import seems_like_discrete_data

        discrete, discrete_values = seems_like_discrete_data(
            numpy.asarray(x).reshape(-1), return_uniques=True
        )

    x_discrete_labels = (
        None if discrete_values is None else [str(i) for i in discrete_values]
    )

    if bins is None:
        if x_discrete_labels is not None:
            # Discrete bins using defined labels
            bins = numpy.arange(len(x_discrete_labels) + 1)
        if isinstance(x.dtype, pandas.CategoricalDtype):
            # Discrete bins using implied labels
            discrete_values = numpy.arange(len(x_discrete_labels))
            bins = numpy.arange(len(x_discrete_labels) + 1)
            x = x.cat.codes
        else:
            x_ = x
            if availability is not None:
                x_ = x[availability != 0]
            low_pctile = 0
            high_pctile = 100
            if range:
                import scipy.stats

                if range[0] is not None:
                    low_pctile = scipy.stats.percentileofscore(x_, range[0])
                if range[1] is not None:
                    high_pctile = scipy.stats.percentileofscore(x_, range[1])
            if isinstance(pct_bins, int):
                bins = numpy.percentile(
                    x_, numpy.linspace(low_pctile, high_pctile, pct_bins + 1)
                )
            else:
                bins = numpy.percentile(x_, pct_bins)
    elif isinstance(bins, int) and availability is not None:
        # Equal width bin generation using only available alternatives
        x_ = x[availability != 0]
        if range:
            range_low, range_high = range
            if range_low is None:
                range_low = x_.min()
            if range_high is None:
                range_high = x_.max()
        else:
            range_low = x_.min()
            range_high = x_.max()
        bins = numpy.linspace(range_low, range_high, bins + 1)

    model_result = probability
    model_choice = choices

    if style == "kde":
        import scipy.stats

        kernel_result = scipy.stats.gaussian_kde(
            x, bw_method=bw_method, weights=model_result.reshape(-1)
        )
        common_bw = kernel_result.covariance_factor()
        if model_choice is not None:
            kernel_choice = scipy.stats.gaussian_kde(
                x, bw_method=common_bw, weights=model_choice.reshape(-1)
            )
        else:
            kernel_choice = None

        if not range:
            x_ = x
            if availability is not None:
                x_ = x[availability != 0]
            range = (x_.min(), x_.max())

        x_points = numpy.linspace(*range, 250)
        y_points_1 = kernel_result(x_points)
        y_points_2 = kernel_choice(x_points)

    else:
        shift = 0.4 if discrete else 0
        gap = 0.2 if discrete else 0
        if range:
            range_low, range_high = range
            if range_low is None:
                range_low = x.min()
            if range_high is None:
                range_high = x.max()
            range = (range_low, range_high)

        y1, x1 = numpy.histogram(
            x,
            weights=model_result.reshape(-1),
            bins=bins,
            range=range,
            density=True,
        )
        x_points, y_points_1 = pseudo_bar_data(x1 - shift, y1, gap=gap)

        if model_choice is not None:
            y2, x2 = numpy.histogram(
                x,
                weights=model_choice.reshape(-1),
                bins=x1,
                density=True,
            )
            x_points, y_points_2 = pseudo_bar_data(x1 - shift, y2, gap=gap)

            if coincidence_ratio:
                _coincidence_ratio = (
                    numpy.minimum(y1, y2).sum() / numpy.maximum(y1, y2).sum()
                )

    if xlabel is False:
        xlabel = None

    if accumulator and model_choice is not None:
        fig, (ax, ax2) = plt.subplots(
            2, 1, gridspec_kw={"height_ratios": [3, 1], "hspace": 0}, sharex="col"
        )
        cum_sum = (y1 - y2).cumsum()
        x1plus = (x1[:-1] + x1[1:]) / 2
        ax2.plot(x1plus, cum_sum, color="k", lw=0.5)
        ax2.fill_between(
            x1plus,
            numpy.fmin(cum_sum, 0),
            0,
            facecolor="#ffa200",
            label=f"{prob_label} Ahead",
        )
        ax2.fill_between(
            x1plus,
            numpy.fmax(cum_sum, 0),
            0,
            facecolor="#1f77b4",
            label=f"{obs_label} Ahead",
        )
        ax2.set_yticks([])
        ax2.set_ylabel("Net Cum.")
        ax2.legend()

    elif ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    if _coincidence_ratio:
        ax.text(
            0.5,
            0.98,
            f"Coincidence Ratio = {_coincidence_ratio:0.4f}",
            horizontalalignment="center",
            verticalalignment="top",
            transform=ax.transAxes,
        )

    ax.bins = bins
    ax.plot(x_points, y_points_1, label=prob_label, lw=1.5)
    if model_choice is not None:
        ax.fill_between(
            x_points,
            y_points_2,
            label=obs_label,
            step=None,
            facecolor="#ffbe4d",
            edgecolor="#ffa200",
            lw=1.5,
        )
    ax.legend()
    if not discrete:
        ax.set_xlim(x_points[0], x_points[-1])
        if xscale:
            if isinstance(xscale, str):
                ax.set_xscale(xscale)
            elif isinstance(xscale, dict):
                ax.set_xscale(**xscale)
            else:
                raise ValueError(f"xscale must be str or dict, not {type(xscale)}")
    if xmajorticks is not None:
        ax.set_xticks(xmajorticks)
        ax.set_xticklabels(xmajorticks)
    if xminorticks is not None:
        ax.set_xticks(xminorticks, minor=True)

    if x_discrete_labels is not None:
        ax.set_xticks(numpy.arange(len(x_discrete_labels)))
        ax.set_xticklabels(x_discrete_labels)
    if accumulator and model_choice is not None:
        ax2.set_xlabel(xlabel)
    else:
        ax.set_xlabel(xlabel)
    ax.set_yticks([])
    ax.set_ylabel(ylabel)
    if fig is None:
        return ax
    fig.tight_layout(pad=0.5)
    if format == "svg":
        result = plot_as_svg_xhtml(fig, **kwargs)
        fig.clf()
        plt.close(fig)
    elif format == "png":
        from .png import make_png

        result = make_png(fig, **kwargs)
        fig.clf()
        plt.close(fig)
    else:
        result = ax
    return result


def histogram_on_idca_variable(
    x,
    pr,  # array-like
    *,
    ds: Dataset = None,
    dt: DataTree = None,
    bins=10,
    x_label=None,
    y_label="Relative Frequency",
    prob_label="Modeled",
    obs_label="Observed",
    span=None,
    filter_co=None,
):
    """
    Generate a figure of observed and modeled choices over a range of variable values.

    Parameters
    ----------
    x : str or array-like
    pr
    ds : Dataset
    dt : DataTree
    bins : int, optional
    x_label : str, optional
    y_label : str, default "Relative Frequency"
    prob_label : str, default "Modeled"
    obs_label : str, default "Observed"
    span : tuple, optional
    filter_co : str or array-like, optional

    Returns
    -------
    altair.Chart
    """
    import altair as alt

    ch = None
    av = None

    if isinstance(x, str):
        x_label = x_label or x
        if ds is not None and "ca" in ds and x in ds["var_ca"]:
            x = ds["ca"].sel(var_ca=x).values.reshape(-1)
        if isinstance(x, str) and dt is not None:
            x = dt.get_expr(x, allow_native=False).values.reshape(-1)

    if isinstance(x, str):
        raise ValueError("x not decoded")

    # n_cases = pr.shape[0]
    n_alts = pr.shape[1]

    if ds is not None:
        if "ch" in ds:
            ch = ds["ch"].values[:, :n_alts].reshape(-1)
        if "av" in ds:
            av = (ds["av"].values[:, :n_alts] != 0).reshape(-1)
    else:
        raise NotImplementedError

    if isinstance(bins, int):
        # Equal width bin generation using only available alternatives
        x_ = x[av]
        if span is not None:
            range_low, range_high = span
            if range_low is None:
                range_low = x_.min()
            if range_high is None:
                range_high = x_.max()
        else:
            range_low = x_.min()
            range_high = x_.max()
        bins = numpy.linspace(range_low, range_high, bins + 1)

    if filter_co is not None:
        if isinstance(filter_co, str):
            if ds is not None and "co" in ds and filter_co in ds["var_co"]:
                filter_co = ds["co"].sel(var_co=filter_co).values.astype(bool)
            if isinstance(filter_co, str) and dt is not None:
                datatree_co = dt.idco_subtree()
                filter_co = datatree_co.get_expr(
                    filter_co, allow_native=False
                ).values.astype(bool)
        filter_shape = tuple(filter_co.shape) + (-1,)
        x = numpy.asarray(x).reshape(*filter_shape)[filter_co].reshape(-1)
        pr = pr.reshape(*filter_shape)[filter_co].reshape(-1)
        ch = ch.reshape(*filter_shape)[filter_co].reshape(-1)

    y_points_1, x1 = numpy.histogram(
        x,
        weights=pr.reshape(x.shape),
        bins=bins,
        density=True,
    )

    y_points_2, x2 = numpy.histogram(
        x,
        weights=ch.reshape(x.shape),
        bins=x1,
        density=True,
    )

    d1 = {
        x_label: x1,
        y_label: np.append(y_points_1, 0),
        "source": prob_label,
        "strokeOpacity": 1,
        "fillOpacity": 0,
    }
    d2 = {
        x_label: x1,
        y_label: np.append(y_points_2, 0),
        "source": obs_label,
        "strokeOpacity": 0,
        "fillOpacity": 1,
    }

    fig_data = pd.concat(
        [
            pd.DataFrame(d1),
            pd.DataFrame(d2),
        ]
    )

    return (
        alt.Chart(fig_data)
        .mark_area(
            interpolate="step-after",
            line=True,
        )
        .encode(
            x=x_label,
            y=alt.Y(y_label, axis=alt.Axis(labels=False)),
            color="source",
            fillOpacity=alt.FillOpacity("fillOpacity", scale=None),
            strokeOpacity=alt.StrokeOpacity("strokeOpacity", scale=None),
        )
    )


def distribution_on_idca_variable(
    model,
    x,
    xlabel=None,
    ylabel="Relative Frequency",
    style="hist",
    bins=None,
    pct_bins=20,
    range=None,
    xlim=None,
    prob_label="Modeled",
    obs_label="Observed",
    subselector=None,
    probability=None,
    bw_method=None,
    discrete=None,
    ax=None,
    format="ax",
    **kwargs,
):
    """
    Generate a figure of observed and modeled choices over a range of variable values.

    Parameters
    ----------
    model : BaseModel
            The discrete choice model to analyze.
    x : str or array-like
            The name of an `idca` variable, or an array giving its values.  If this name exactly
            matches that of an `idca` column in the model's loaded `dataset`, then
            those values are used, otherwise the variable is loaded from the model's
            `dataservice`.
    xlabel : str, optional
            A label to use for the x-axis of the resulting figure.  If not given,
            the value of `x` is used if it is a string.  Set to `False` to omit the
            x-axis label.
    ylabel : str, default "Relative Frequency"
            A label to use for the y-axis of the resulting figure.
    style : {'hist', 'kde'}
            The style of figure to produce, either a histogram or a kernel density plot.
    bins : int, default 25
            The number of bins to use, only applicable to histogram style.
    range : 2-tuple, optional
            A range to truncate the figure. (alias `xlim`)
    prob_label : str, default "Modeled"
            A label to put in the legend for the modeled probabilities
    obs_label : str, default "Observed"
            A label to put in the legend for the observed choices
    subselector : str or array-like, optional
            A filter to apply to cases. If given as a string, this is loaded from the
            model's `dataservice` as an `idco` variable.
    probability : array-like, optional
            The pre-calculated probability array for all cases in this analysis.
            If not given, the probability array is calculated at the current parameter
            values.
    ax : matplotlib.Axes, optional
            If given, the figure will be drawn on these axes and they will be returned,
            otherwise new blank axes are used to draw the figure.
    format : {'ax', 'figure', 'svg'}, default 'figure'
            How to return the result if it is a figure. The default is to return
            the raw matplotlib Axes instance.  Change this to `svg` to get a SVG
            rendering as an xmle.Elem.

    Other Parameters
    ----------------
    header : str, optional
            A header to attach to the figure.  The header is not generated using
            matplotlib, but instead is prepended to the xml output with a header tag before the
            rendered svg figure.


    Returns
    -------
    Elem or Axes
            Returns `ax` if given as an argument, otherwise returns a rendering as an Elem
    """
    if model is None:
        return lambda mdl: distribution_on_idca_variable(
            mdl,
            xlabel=xlabel,
            ylabel=ylabel,
            style=style,
            bins=bins,
            pct_bins=pct_bins,
            range=range,
            xlim=xlim,
            prob_label=prob_label,
            obs_label=obs_label,
            subselector=subselector,
            probability=probability,
            bw_method=bw_method,
            discrete=discrete,
            ax=ax,
            format=format,
            **kwargs,
        )

    if xlim is not None and range is None:
        range = xlim

    if isinstance(x, str):
        x_label = x
        if model.dataset and "ca" in model.dataset and x in model.dataset["var_ca"]:
            x = model.dataset["ca"][x].values.reshape(-1)
        # elif model.dataservice is not None:
        #     x = (
        #         model.dataservice.make_dataframes({"ca": [x]}, explicit=True)
        #         .array_ca()
        #         .reshape(-1)
        #     )
        # elif getattr(model, "datatree", None) is not None:
        #     x = model.datatree.get_expr(x).values
        else:
            raise ValueError("model is missing data source")
    else:
        try:
            x_label = x.name
        except AttributeError:
            x_label = ""

    # if model.dataframes \
    #       and model.dataframes.data_ca is not None \
    #       and continuous_variable in model.dataframes.data_ca:
    # 	cv = model.dataframes.data_ca[continuous_variable].values.reshape(-1)
    # else:
    # 	cv = model.dataservice.make_dataframes(
    #   	{'ca': [continuous_variable]}, explicit=True
    #   ).array_ca().reshape(-1)

    discrete_values = None
    if discrete:
        discrete_values = numpy.unique(x)
    elif discrete is None:
        from .histograms import seems_like_discrete_data

        discrete, discrete_values = seems_like_discrete_data(
            numpy.asarray(x).reshape(-1), return_uniques=True
        )

    x_discrete_labels = (
        None if discrete_values is None else [str(i) for i in discrete_values]
    )

    if bins is None:
        if x_discrete_labels is not None:
            # Discrete bins using defined labels
            bins = numpy.arange(len(x_discrete_labels) + 1)
        if isinstance(x.dtype, pandas.CategoricalDtype):
            # Discrete bins using implied labels
            discrete_values = numpy.arange(len(x_discrete_labels))
            bins = numpy.arange(len(x_discrete_labels) + 1)
            x = x.cat.codes
        else:
            x_ = x
            if model.dataset is not None:
                if "av" in model.dataset and model.dataframes.data_ca is not None:
                    x_ = x[model.dataset["av"].values.reshape(-1) != 0]
            elif getattr(model, "datatree", None) is not None:
                if model.availability_ca_var or model.availability_co_vars:
                    raise NotImplementedError
            low_pctile = 0
            high_pctile = 100
            if range is not None:
                import scipy.stats

                if range[0] is not None:
                    low_pctile = scipy.stats.percentileofscore(x_, range[0])
                if range[1] is not None:
                    high_pctile = scipy.stats.percentileofscore(x_, range[1])
            if isinstance(pct_bins, int):
                bins = numpy.percentile(
                    x_, numpy.linspace(low_pctile, high_pctile, pct_bins + 1)
                )
            else:
                bins = numpy.percentile(x_, pct_bins)
    elif (
        isinstance(bins, int)
        and model.dataset is not None
        and "av" in model.dataset
        and "ca" in model.dataset
    ):
        # Equal width bin generation using only available alternatives
        x_ = x[model.dataset["av"].values.reshape(-1) != 0]
        if range is not None:
            range_low, range_high = range
            if range_low is None:
                range_low = x_.min()
            if range_high is None:
                range_high = x_.max()
        else:
            range_low = x_.min()
            range_high = x_.max()
        bins = numpy.linspace(range_low, range_high, bins + 1)

    if probability is None:
        probability = model.probability()

    if model.dataset is not None:
        n_alts = model.dataset.dc.n_alts
    elif getattr(model, "datatree", None) is not None:
        n_alts = model.datatree.n_alts
    else:
        raise ValueError("model is missing data source")
    model_result = probability[:, :n_alts]
    if model.dataset is not None:
        model_choice = model.dataset["ch"].values
        if "wt" in model.dataset:
            model_wt = model.dataset["wt"]
        else:
            model_wt = None
    else:
        model_choice = model.data_as_loaded["ch"].values
        if "wt" in model.data_as_loaded:
            model_wt = model.data_as_loaded["wt"]
        else:
            model_wt = None

    if model_wt is not None:
        model_result = model_result.copy()
        model_result *= model_wt.values.reshape(-1, 1)
        model_choice = model_choice.copy()
        model_choice *= model_wt.values.reshape(-1, 1)

    if subselector is not None:
        if isinstance(subselector, str):
            if getattr(model, "datatree", None) is not None:
                subselector = model.datatree.get_expr(subselector).values.astype(bool)
        x = numpy.asarray(x).reshape(*model_result.shape)[subselector].reshape(-1)
        model_result = model_result[subselector]
        model_choice = model_choice[subselector]

    if style == "kde":
        import scipy.stats

        kernel_result = scipy.stats.gaussian_kde(
            x.reshape(-1), bw_method=bw_method, weights=model_result.reshape(-1)
        )
        common_bw = kernel_result.covariance_factor()
        kernel_choice = scipy.stats.gaussian_kde(
            x.reshape(-1), bw_method=common_bw, weights=model_choice.reshape(-1)
        )

        if range is None:
            x_ = x
            if model.dataframes is not None:
                if (
                    model.dataframes.data_av is not None
                    and model.dataframes.data_ca is not None
                ):
                    x_ = x[model.dataframes.data_av.values.reshape(-1) != 0]
                elif getattr(model, "datatree", None) is not None:
                    if model.availability_ca_var or model.availability_co_vars:
                        raise NotImplementedError
            range = (x_.min(), x_.max())

        x_points = numpy.linspace(*range, 250)
        y_points_1 = kernel_result(x_points)
        y_points_2 = kernel_choice(x_points)

    else:
        if range is not None:
            range_low, range_high = range
            if range_low is None:
                range_low = x.min()
            if range_high is None:
                range_high = x.max()
            range = (range_low, range_high)

        y_points_1, x1 = numpy.histogram(
            x,
            weights=model_result.reshape(x.shape),
            bins=bins,
            range=range,
            density=True,
        )

        y_points_2, x2 = numpy.histogram(
            x,
            weights=model_choice.reshape(x.shape),
            bins=x1,
            density=True,
        )

        shift = 0.4 if discrete else 0
        gap = 0.2 if discrete else 0

        x_points, y_points_1 = pseudo_bar_data(x1 - shift, y_points_1, gap=gap)
        x_points, y_points_2 = pseudo_bar_data(x1 - shift, y_points_2, gap=gap)

    if xlabel is None:
        xlabel = x_label
    if xlabel is False:
        xlabel = None

    if ax is None:
        fig, ax = plt.subplots()
    else:
        fig = None

    ax.bins = bins
    ax.plot(x_points, y_points_1, label=prob_label, lw=1.5)
    ax.fill_between(
        x_points,
        y_points_2,
        label=obs_label,
        step=None,
        facecolor="#ffbe4d",
        edgecolor="#ffa200",
        lw=1.5,
    )
    ax.legend()
    if not discrete:
        ax.set_xlim(x_points[0], x_points[-1])
    if x_discrete_labels is not None:
        ax.set_xticks(numpy.arange(len(x_discrete_labels)))
        ax.set_xticklabels(x_discrete_labels)
    ax.set_xlabel(xlabel)
    ax.set_yticks([])
    ax.set_ylabel(ylabel)
    if fig is None:
        return ax
    fig.tight_layout(pad=0.5)
    if format == "svg":
        result = plot_as_svg_xhtml(fig, **kwargs)
        fig.clf()
        plt.close(fig)
    elif format == "png":
        from .png import make_png

        result = make_png(fig, **kwargs)
        fig.clf()
        plt.close(fig)
    else:
        result = ax
    return result


BaseModel.distribution_on_idca_variable = distribution_on_idca_variable


def share_figure(
    x,
    probability,
    choices=None,
    weights=1,
    xlabel=None,
    bins=None,
    pct_bins=20,
    figsize=(12, 4),
    style="stacked",
    discrete=None,
    xlim=None,
    xscale=None,
    xmajorticks=None,
    xminorticks=None,
    include_nests=False,
    exclude_alts=None,
    format="figure",
    **kwargs,
):
    """
    Generate a figure of variables over a range of variable values.

    Parameters
    ----------
    x : array-like, 1-d
            An array giving values for some variable.
    probability : array-like, 2-d
            The pre-calculated probability array for all cases in this analysis.
            First dimension must be the same shape as `x`.  The second dimension
            represents the alternatives (or similar).
    choices : array-like, optional
            The observed choices array for all cases in this analysis. If provided,
            the first dimension must be the same shape as `x`.  The second dimension
            represents the alternatives (or similar).
    weights : array-like, 1-d, optional
            The case weights for all cases in this analysis. If provided,
            the shape must be the same shape as `x`.
    xlabel : str, optional
            A label to use for the x-axis of the resulting figure.  If not given,
            the value of `x.name` is used if it exists.  Set to `False` to omit the
            x-axis label.
    bins : int, optional
            The number of equal-sized bins to use.
    pct_bins : int or array-like, default 20
            The number of equal-mass bins to use.
    style : {'stacked', 'dataframe', 'many'}
            The type of output to generate.
    discrete : bool, default False
            Whether to treat the data values explicitly as discrete (vs continuous)
            data.  This will change the styling and automatic bin generation.  If
            there are very few unique values, the data will be assumed to be
            discrete anyhow.
    xlim : 2-tuple, optional
            Explicitly set the range of values shown on the x axis of generated
            figures.  This can truncate long tails.  The actual histogram bins
            are not changed.
    include_nests : bool, default False
            Whether to include nests in the figure.
    exclude_alts : Collection, optional
            Alternatives to exclude from the figure.
    filter : str, optional
            A filter that will be used to select only a subset of cases.
    format : {'figure','svg'}, default 'figure'
            How to return the result if it is a figure. The default is to return
            the raw matplotlib Figure instance, ot set to `svg` to get a SVG
            rendering as an xmle.Elem.

    Returns
    -------
    Figure, DataFrame, or Elem
    """
    if style not in {"stacked", "dataframe", "many"}:
        raise ValueError("style must be in {'stacked', 'dataframe', 'many'}")

    if include_nests and style == "stacked" and exclude_alts is None:
        import warnings

        warnings.warn(
            "including nests in a stacked figure is likely to give "
            "misleading results unless constituent alternatives are omitted",
            stacklevel=2,
        )

    if exclude_alts is None:
        exclude_alts = set()

    if xlabel is None:
        try:
            xlabel = x.name
        except AttributeError:
            pass

    h_pr = {}
    h_ch = {}

    discrete_values = None
    if discrete:
        discrete_values = numpy.unique(x)
    elif discrete is None:
        from .histograms import seems_like_discrete_data

        discrete, discrete_values = seems_like_discrete_data(x, return_uniques=True)

    pr = numpy.asarray(probability)
    if choices is not None:
        ch = numpy.asarray(choices)
    else:
        ch = None
    wt = numpy.asarray(weights)

    x_discrete_labels = (
        None if discrete_values is None else [str(i) for i in discrete_values]
    )

    if bins is None:
        if isinstance(x.dtype, pandas.CategoricalDtype):
            discrete_values = numpy.arange(len(x_discrete_labels))
            bins = numpy.arange(len(x_discrete_labels) + 1)
            x = x.cat.codes
        elif isinstance(pct_bins, int):
            bins = numpy.percentile(x, numpy.linspace(0, 100, pct_bins + 1))
        else:
            bins = numpy.percentile(x, pct_bins)

    try:
        columns = probability.columns
    except AttributeError:
        columns = None
    else:
        columns = dict(enumerate(columns))

    # check for correct array shapes, raise helpful message if not compatible
    pr_w_shape = numpy.broadcast_shapes(pr[:, 0].shape, wt.shape)
    if x.shape != pr_w_shape:
        raise ValueError(
            f"incompatible shapes, "
            f"x.shape={x.shape}, "
            f"pr.shape={pr.shape}, "
            f"wt.shape={wt.shape}, "
            f"(pr[:,i]*wt).shape={pr_w_shape}"
        )
    if ch is not None:
        ch_w_shape = numpy.broadcast_shapes(ch[:, 0].shape, wt.shape)
        if x.shape != ch_w_shape:
            raise ValueError(
                f"incompatible shapes, "
                f"x.shape={x.shape}, "
                f"ch.shape={ch.shape}, "
                f"wt.shape={wt.shape}, "
                f"(ch[:,i]*wt).shape={ch_w_shape}"
            )

    for i in range(pr.shape[1]):
        h_pr[i], _ = numpy.histogram(
            x,
            weights=pr[:, i] * wt,
            bins=bins,
        )
        if ch is not None:
            h_ch[i], _ = numpy.histogram(
                x,
                weights=ch[:, i] * wt,
                bins=bins,
            )

    h_pr = pandas.DataFrame(h_pr)
    h_pr.index = pandas.IntervalIndex.from_breaks(bins)  # bins[:-1]
    h_pr.rename(columns=columns, inplace=True)
    _denominator, _ = numpy.histogram(
        x,
        weights=pr.sum(1) * wt,
        bins=bins,
    )
    h_pr_share = h_pr / _denominator.reshape(-1, 1)
    if ch is not None:
        _denominator_ch, _ = numpy.histogram(
            x,
            weights=ch.sum(1) * wt,
            bins=bins,
        )
        h_ch = pandas.DataFrame(h_ch)
        h_ch.index = h_pr.index
        h_ch.rename(columns=columns, inplace=True)
        h_ch_share = h_ch / _denominator_ch.reshape(-1, 1)
    else:
        h_ch_share = None

    if discrete:
        x_placement = numpy.arange(len(bins) - 1)
        x_alignment = "center"
        bin_widths = 0.8
    else:
        x_placement = bins[:-1]
        x_alignment = "edge"
        bin_widths = bins[1:] - bins[:-1]

    if xlabel is False:
        xlabel = None

    if xlim is None:
        xlim = (bins[0], bins[-1])

    if style == "dataframe":
        if ch is not None:
            result = pandas.concat(
                {
                    "Modeled Shares": h_pr_share,
                    "Observed Shares": h_ch_share,
                },
                axis=1,
                sort=False,
            )
        else:
            result = pandas.concat(
                {
                    "Modeled Shares": h_pr_share,
                },
                axis=1,
                sort=False,
            )
        result["Count", "*"] = h_pr.sum(1)

        if xlabel:
            result.index.name = xlabel

    elif style == "stacked":
        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=figsize)

        bottom0 = 0
        bottom1 = 0

        for i in h_pr_share.columns:
            ax0.bar(
                x_placement,
                height=h_pr_share[i],
                bottom=bottom0,
                width=bin_widths,
                align=x_alignment,
                label=i,
            )
            bottom0 = h_pr_share[i].fillna(0).values + bottom0
            ax1.bar(
                x_placement,
                height=h_ch_share[i],
                bottom=bottom1,
                width=bin_widths,
                align=x_alignment,
            )
            bottom1 = h_ch_share[i].fillna(0).values + bottom1

        ax0.set_ylim(0, 1)
        if not discrete:
            ax0.set_xlim(*xlim)
            if xscale:
                if isinstance(xscale, str):
                    ax0.set_xscale(xscale)
                elif isinstance(xscale, dict):
                    ax0.set_xscale(**xscale)
                else:
                    raise ValueError(f"xscale must be str or dict, not {type(xscale)}")
            if xmajorticks is not None:
                ax0.set_xticks(xmajorticks)
                ax0.set_xticklabels(xmajorticks)
            if xminorticks is not None:
                ax0.set_xticks(xminorticks, minor=True)
        if x_discrete_labels is not None:
            ax0.set_xticks(numpy.arange(len(x_discrete_labels)))
            ax0.set_xticklabels(x_discrete_labels)
        ax0.set_title("Modeled Shares")

        ax1.set_ylim(0, 1)
        if not discrete:
            ax1.set_xlim(*xlim)
            if xscale:
                if isinstance(xscale, str):
                    ax1.set_xscale(xscale)
                elif isinstance(xscale, dict):
                    ax1.set_xscale(**xscale)
                else:
                    raise ValueError(f"xscale must be str or dict, not {type(xscale)}")
            if xmajorticks is not None:
                ax1.set_xticks(xmajorticks)
                ax1.set_xticklabels(xmajorticks)
            if xminorticks is not None:
                ax1.set_xticks(xminorticks, minor=True)
        if x_discrete_labels is not None:
            ax1.set_xticks(numpy.arange(len(x_discrete_labels)))
            ax1.set_xticklabels(x_discrete_labels)
        ax1.set_title("Observed Shares")
        if xlabel:
            ax0.set_xlabel(xlabel)
            ax1.set_xlabel(xlabel)

        fig.legend(
            loc="center right",
        )

        # fig.tight_layout(pad=0.5)
        if format == "svg":
            result = plot_as_svg_xhtml(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        elif format == "png":
            from .png import make_png

            result = make_png(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        else:
            result = fig

    else:
        fig, axes = plt.subplots(len(h_pr_share.columns), 1, figsize=figsize)

        shift = 0.4 if discrete else 0

        for n, i in enumerate(h_pr_share.columns):
            x_, y_ = pseudo_bar_data(
                bins - shift, h_pr_share[i], gap=0.2 if discrete else 0
            )
            axes[n].plot(x_, y_, label="Modeled" if n == 0 else None, lw=1.5)

            x_ch_, y_ch_ = pseudo_bar_data(
                bins - shift, h_ch_share[i], gap=0.2 if discrete else 0
            )
            axes[n].fill_between(
                x_ch_,
                y_ch_,
                label="Observed" if n == 0 else None,
                step=None,
                facecolor="#ffbe4d",
                edgecolor="#ffa200",
                lw=1.5,
            )
            if not discrete:
                axes[n].set_xlim(*xlim)
                if xscale:
                    if isinstance(xscale, str):
                        axes[n].set_xscale(xscale)
                    elif isinstance(xscale, dict):
                        axes[n].set_xscale(**xscale)
                    else:
                        raise ValueError(
                            f"xscale must be str or dict, not {type(xscale)}"
                        )
                if xmajorticks is not None:
                    axes[n].set_xticks(xmajorticks)
                    axes[n].set_xticklabels(xmajorticks)
                if xminorticks is not None:
                    axes[n].set_xticks(xminorticks, minor=True)
            if x_discrete_labels is not None:
                axes[n].set_xticks(numpy.arange(len(x_discrete_labels)))
                axes[n].set_xticklabels(x_discrete_labels)
            axes[n].set_ylabel(i)

            # axes[n].legend(
            # 	# loc='center right',
            # )

        legnd = axes[0].legend(
            loc="lower center", ncol=2, borderaxespad=0, bbox_to_anchor=(0.5, 1.08)
        )

        if xlabel:
            axes[-1].set_xlabel(xlabel)
        # fig.tight_layout(pad=0.5)
        if format == "svg":
            result = plot_as_svg_xhtml(fig, bbox_extra_artists=[legnd], **kwargs)
            fig.clf()
            plt.close(fig)
        elif format == "png":
            from .png import make_png

            result = make_png(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        else:
            result = fig

    return result


def distribution_on_idco_variable(
    model,
    x,
    xlabel=None,
    bins=None,
    pct_bins=20,
    figsize=(12, 4),
    style="stacked",
    discrete=None,
    xlim=None,
    include_nests=False,
    exclude_alts=None,
    filter=None,
    format="figure",
    **kwargs,
):
    """
    Generate a figure of variables over a range of variable values.

    Parameters
    ----------
    model : BaseModel
            The discrete choice model to analyze.
    x : str or array-like
            The name of an `idco` variable, or an array giving its values.  If this name exactly
            matches that of an `idco` column in the model's loaded `dataframes`, then
            those values are used, otherwise the variable is loaded from the model's
            `dataservice`.
    xlabel : str, optional
            A label to use for the x-axis of the resulting figure.  If not given,
            the value of `x` is used if it is a string.  Set to `False` to omit the
            x-axis label.
    bins : int, optional
            The number of equal-sized bins to use.
    pct_bins : int or array-like, default 20
            The number of equal-mass bins to use.
    style : {'stacked', 'dataframe', 'many'}
            The type of output to generate.
    discrete : bool, default False
            Whether to treat the data values explicitly as discrete (vs continuous)
            data.  This will change the styling and automatic bin generation.  If
            there are very few unique values, the data will be assumed to be
            discrete anyhow.
    xlim : 2-tuple, optional
            Explicitly set the range of values shown on the x axis of generated
            figures.  This can truncate long tails.  The actual histogram bins
            are not changed.
    include_nests : bool, default False
            Whether to include nests in the figure.
    exclude_alts : Collection, optional
            Alternatives to exclude from the figure.
    filter : str, optional
            A filter that will be used to select only a subset of cases.
    format : {'figure','svg'}, default 'figure'
            How to return the result if it is a figure. The default is to return
            the raw matplotlib Figure instance, ot set to `svg` to get a SVG
            rendering as an xmle.Elem.

    Returns
    -------
    Figure, DataFrame, or Elem
    """
    if style not in {"stacked", "dataframe", "many"}:
        raise ValueError("style must be in {'stacked', 'dataframe', 'many'}")

    if include_nests and style == "stacked" and exclude_alts is None:
        import warnings

        warnings.warn(
            "including nests in a stacked figure is likely to give "
            "misleading results unless constituent alternatives are omitted",
            stacklevel=2,
        )

    if exclude_alts is None:
        exclude_alts = set()

    if isinstance(x, str):
        x_label = x
        if (
            model.dataframes
            and model.dataframes.data_co is not None
            and x in model.dataframes.data_co
        ):
            x = model.dataframes.data_co[x].values.reshape(-1)
        else:
            x = (
                model.dataservice.make_dataframes({"co": [x]}, explicit=True)
                .array_co()
                .reshape(-1)
            )
    else:
        try:
            x_label = x.name
        except AttributeError:
            x_label = ""

    if filter:
        _ds = model.dataservice if model.dataservice is not None else model.dataframes
        filter_ = (
            _ds.make_dataframes(
                {"co": [filter]},
                explicit=True,
                float_dtype=bool,
            )
            .array_co()
            .reshape(-1)
        )
        x = x[filter_]
    else:
        filter_ = slice(None)

    h_pr = {}
    h_ch = {}

    discrete_values = None
    if discrete:
        discrete_values = numpy.unique(x)
    elif discrete is None:
        from .histograms import seems_like_discrete_data

        discrete, discrete_values = seems_like_discrete_data(x, return_uniques=True)

    pr = model.probability(
        return_dataframe="names",
        include_nests=bool(include_nests),
    ).loc[filter_, :]

    if include_nests:
        ch = model.dataframes.data_ch_cascade(model.graph).loc[filter_, :]
    else:
        ch = model.dataframes.data_ch.loc[filter_, :]

    if model.dataframes.data_wt is None:
        wt = 1
    else:
        wt = model.dataframes.data_wt.values.reshape(-1)[filter_]

    x_discrete_labels = (
        None if discrete_values is None else [str(i) for i in discrete_values]
    )

    if bins is None:
        if isinstance(x.dtype, pandas.CategoricalDtype):
            discrete_values = numpy.arange(len(x_discrete_labels))
            bins = numpy.arange(len(x_discrete_labels) + 1)
            x = x.cat.codes
        elif isinstance(pct_bins, int):
            bins = numpy.percentile(x, numpy.linspace(0, 100, pct_bins + 1))
        else:
            bins = numpy.percentile(x, pct_bins)

    n_alts = model.graph.n_elementals()
    columns = {}

    for i in range(pr.shape[1]):
        columns[i] = pr.columns[i]
        if (
            i < n_alts
            or include_nests is True
            or model.graph.standard_sort[i] in include_nests
        ):
            if model.graph.standard_sort[i] == model.graph.root_id:
                continue
            if model.graph.standard_sort[i] in exclude_alts:
                continue
            h_pr[i], _ = numpy.histogram(
                x,
                weights=pr.iloc[:, i] * wt,
                bins=bins,
            )
            h_ch[i], _ = numpy.histogram(
                x,
                weights=ch.iloc[:, i] * wt,
                bins=bins,
            )

    h_pr = pandas.DataFrame(h_pr)
    h_pr.index = pandas.IntervalIndex.from_breaks(bins)  # bins[:-1]
    h_pr.rename(columns=columns, inplace=True)
    _denominator, _ = numpy.histogram(
        x,
        weights=numpy.ones_like(pr.iloc[:, -1]) * wt,
        bins=bins,
    )
    h_pr_share = h_pr / _denominator.reshape(-1, 1)
    h_ch = pandas.DataFrame(h_ch)
    h_ch.index = h_pr.index
    h_ch.rename(columns=columns, inplace=True)
    h_ch_share = h_ch / _denominator.reshape(-1, 1)

    if discrete:
        x_placement = numpy.arange(len(bins) - 1)
        x_alignment = "center"
        bin_widths = 0.8
    else:
        x_placement = bins[:-1]
        x_alignment = "edge"
        bin_widths = bins[1:] - bins[:-1]

    if xlabel is None:
        xlabel = x_label
    if xlabel is False:
        xlabel = None

    if xlim is None:
        xlim = (bins[0], bins[-1])

    if style == "dataframe":
        result = pandas.concat(
            {
                "Modeled Shares": h_pr_share,
                "Observed Shares": h_ch_share,
            },
            axis=1,
            sort=False,
        )
        result["Count", "*"] = h_pr.sum(1)

        if x_label:
            result.index.name = xlabel

    elif style == "stacked":
        fig, (ax0, ax1) = plt.subplots(1, 2, figsize=figsize)

        bottom0 = 0
        bottom1 = 0

        for i in h_pr_share.columns:
            ax0.bar(
                x_placement,
                height=h_pr_share[i],
                bottom=bottom0,
                width=bin_widths,
                align=x_alignment,
                label=i,
            )
            bottom0 = h_pr_share[i].values + bottom0
            ax1.bar(
                x_placement,
                height=h_ch_share[i],
                bottom=bottom1,
                width=bin_widths,
                align=x_alignment,
            )
            bottom1 = h_ch_share[i].values + bottom1

        ax0.set_ylim(0, 1)
        if not discrete:
            ax0.set_xlim(*xlim)
        if x_discrete_labels is not None:
            ax0.set_xticks(numpy.arange(len(x_discrete_labels)))
            ax0.set_xticklabels(x_discrete_labels)
        ax0.set_title("Modeled Shares")

        ax1.set_ylim(0, 1)
        if not discrete:
            ax1.set_xlim(*xlim)
        if x_discrete_labels is not None:
            ax1.set_xticks(numpy.arange(len(x_discrete_labels)))
            ax1.set_xticklabels(x_discrete_labels)
        ax1.set_title("Observed Shares")
        if xlabel:
            ax0.set_xlabel(xlabel)
            ax1.set_xlabel(xlabel)

        fig.legend(
            loc="center right",
        )

        # fig.tight_layout(pad=0.5)
        if format == "svg":
            result = plot_as_svg_xhtml(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        elif format == "png":
            from .png import make_png

            result = make_png(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        else:
            result = fig

    else:
        fig, axes = plt.subplots(len(h_pr_share.columns), 1, figsize=figsize)

        shift = 0.4 if discrete else 0

        for n, i in enumerate(h_pr_share.columns):
            x_, y_ = pseudo_bar_data(
                bins - shift, h_pr_share[i], gap=0.2 if discrete else 0
            )
            axes[n].plot(x_, y_, label="Modeled" if n == 0 else None, lw=1.5)

            x_ch_, y_ch_ = pseudo_bar_data(
                bins - shift, h_ch_share[i], gap=0.2 if discrete else 0
            )
            axes[n].fill_between(
                x_ch_,
                y_ch_,
                label="Observed" if n == 0 else None,
                step=None,
                facecolor="#ffbe4d",
                edgecolor="#ffa200",
                lw=1.5,
            )
            if not discrete:
                axes[n].set_xlim(*xlim)
            if x_discrete_labels is not None:
                axes[n].set_xticks(numpy.arange(len(x_discrete_labels)))
                axes[n].set_xticklabels(x_discrete_labels)
            axes[n].set_ylabel(i)

            # axes[n].legend(
            # 	# loc='center right',
            # )

        legnd = axes[0].legend(
            loc="lower center", ncol=2, borderaxespad=0, bbox_to_anchor=(0.5, 1.08)
        )

        if xlabel:
            axes[-1].set_xlabel(xlabel)
        # fig.tight_layout(pad=0.5)
        if format == "svg":
            result = plot_as_svg_xhtml(fig, bbox_extra_artists=[legnd], **kwargs)
            fig.clf()
            plt.close(fig)
        elif format == "png":
            from .png import make_png

            result = make_png(fig, **kwargs)
            fig.clf()
            plt.close(fig)
        else:
            result = fig

    return result


BaseModel.distribution_on_idco_variable = distribution_on_idco_variable



from collections.abc import MutableMapping
from copy import copy
from types import SimpleNamespace


class Namespace(SimpleNamespace, MutableMapping):
    def __getitem__(self, item):
        try:
            return getattr(self, item)
        except AttributeError:
            raise KeyError(item) from None

    def __setitem__(self, key, value):
        setattr(self, key, value)

    def __delitem__(self, key):
        try:
            delattr(self, key)
        except AttributeError:
            raise KeyError(key) from None

    def __iter__(self):
        return iter(self.__dict__)

    def __len__(self):
        return len(self.__dict__)

    def keys(self):
        return self.__dict__.keys()

    def iloc(self, slices=None, **kwargs):
        result = copy(self)
        if slices is not None:
            kwargs.update(slices)
        for k, v in kwargs.items():
            result[k] = result[k].iloc[v]
        return result



# Errors


class MissingDataError(ValueError):
    pass


class DuplicateColumnNames(ValueError):
    pass


class BHHHSimpleStepFailure(RuntimeError):
    pass


# Warnings


class ParameterNotInModelWarning(UserWarning):
    pass



import logging
from typing import NamedTuple

import numba as nb
import numpy as np

from ..dataset.dim_names import ALTID as _ALTID
from ..dataset.dim_names import CASEALT as _CASEALT
from ..dataset.dim_names import CASEID as _CASEID


class _case_slice:
    def __get__(self, obj, objtype=None):
        self.parent = obj
        return self

    def __getitem__(self, idx):
        kwds = {}
        for k in self.parent._fields:
            attribute = getattr(self.parent, k)
            if attribute is not None and (len(k) == 2 or k == "ce_caseptr"):
                kwds[k] = attribute[idx]
            else:
                kwds[k] = attribute
        return type(self.parent)(**kwds)


class DataArrays(NamedTuple):
    ch: np.ndarray
    av: np.ndarray
    wt: np.ndarray
    co: np.ndarray
    ca: np.ndarray
    ce_data: np.ndarray
    ce_altidx: np.ndarray
    ce_caseptr: np.ndarray

    cs = _case_slice()

    # @property
    # def alternatives(self):
    #     if self.alt_codes is not None:
    #         if self.alt_names is not None:
    #             return dict(zip(self.alt_codes, self.alt_names))
    #         else:
    #             return {i:str(i) for i in self.alt_codes}
    #     else:
    #         raise ValueError("alt_codes not defined")


def to_dataset(dataframes):
    caseindex_name = _CASEID
    altindex_name = _ALTID
    from xarray import DataArray

    from ..dataset import Dataset

    coords = {
        caseindex_name: dataframes.caseindex.values,
        altindex_name: dataframes.alternative_codes(),
    }
    ds = Dataset(coords=coords)
    if dataframes.data_co is not None:
        caseindex_name = dataframes.data_co.index.name
        ds.update(Dataset.from_dataframe(dataframes.data_co))
    if dataframes.data_ca is not None:
        caseindex_name = dataframes.data_ca.index.names[0]
        altindex_name = dataframes.data_ca.index.names[1]
        ds.update(Dataset.from_dataframe(dataframes.data_ca))
    alt_names = dataframes.alternative_names()
    if alt_names:
        ds.coords["alt_names"] = DataArray(alt_names, dims=(altindex_name,))
    return ds


def prepare_data(
    datasource,
    request,
    float_dtype=None,
    cache_dir=None,
    flows=None,
    make_unused_flows=False,
):
    """
    Load data from a DataTree into a computationally-formatted Dataset.

    Parameters
    ----------
    datasource : DataTree or Dataset
    request : Mapping or Model
    float_dtype : dtype, optional
    cache_dir : Path-like, optional
        Directory to cache sharrow flows.
    flows : dict, optional
        Collection of previously prepared flows.

    Returns
    -------
    model_dataset : Dataset
        Ready for computation.
    flows : dict
    """
    log = logging.getLogger("Larch")
    from ..dataset import DataArray, Dataset, DataTree

    if float_dtype is None:
        float_dtype = np.float64
    log.debug(f"building dataset from datashare coords: {datasource.coords}")
    model_dataset = Dataset(
        coords=datasource.coords,
    )
    try:
        if "idcoVars" in datasource.subspaces:
            model_dataset.coords.update(datasource.subspaces["idcoVars"].coords)
    except AttributeError:
        pass
    model_dataset.dc.CASEID = datasource.dc.CASEID
    model_dataset.dc.ALTID = datasource.dc.ALTID

    from .numbamodel import NumbaModel  # avoid circular import

    if isinstance(request, NumbaModel):
        alts = request.graph.elemental_names()
        alt_dim = model_dataset.dc.ALTID or _ALTID
        if model_dataset.dc.ALTID not in model_dataset.coords:
            model_dataset.coords[alt_dim] = DataArray(
                list(alts.keys()), dims=(alt_dim,)
            )
        if "alt_names" not in model_dataset.coords:
            model_dataset.coords["alt_names"] = DataArray(
                list(alts.values()), dims=(alt_dim,)
            )
        request = request.required_data()

    if flows is None:
        flows = {}

    if isinstance(datasource, DataTree):
        log.debug("adopting existing DataTree")
        if not datasource.relationships_are_digitized:
            datasource.digitize_relationships(inplace=True)
        datatree = datasource
        datatree_co = datatree.idco_subtree()
    elif isinstance(datasource, Dataset):
        datatree = datasource.dc.as_tree()
        if not datatree.relationships_are_digitized:
            datatree.digitize_relationships(inplace=True)
        datatree_co = datatree.idco_subtree()
    else:
        log.debug("initializing new DataTree")
        datatree = DataTree(main=datasource)
        datatree.digitize_relationships(inplace=True)
        datatree_co = datatree.idco_subtree()

    if "co" in request:
        log.debug(f"requested co data: {request['co']}")
        model_dataset, flows["co"] = _prep_co(
            model_dataset,
            datatree_co,
            request["co"],
            tag="co",
            dtype=float_dtype,
            cache_dir=cache_dir,
            flow=flows.get("co"),
        )
    if "ca" in request:
        log.debug(f"requested ca data: {request['ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            model_dataset, flows["ca"] = _prep_ca(
                model_dataset,
                datatree,
                request["ca"],
                tag="ca",
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get("ca"),
            )
        else:
            model_dataset, flows["ce"] = _prep_ce(
                model_dataset,
                datatree,
                request["ca"],
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get("ce"),
            )
    if "choice_ca" in request:
        log.debug(f"requested choice_ca data: {request['choice_ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            log.debug("  loading choice_ca data from idca")
            model_dataset, flows["choice_ca"] = _prep_ca(
                model_dataset,
                datatree,
                request["choice_ca"],
                tag="ch",
                preserve_vars=False,
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get("choice_ca"),
                force_flow=make_unused_flows,
            )
        else:
            log.debug("  loading choice_ca data from idce")
            model_dataset, flows["choice_ce"] = _prep_ce(
                model_dataset,
                datatree,
                request["choice_ca"],
                v_tag="choice_ca",
                s_tag="choice_ce",
                preserve_vars=False,
                dtype=float_dtype,
                cache_dir=cache_dir,
                flow=flows.get("choice_ce"),
                attach_indexes=False,
                # force_flow=make_unused_flows,
            )
            da_ch = DataArray(
                ce_to_dense(
                    model_dataset["choice_ce_data"].values,
                    model_dataset[model_dataset.dc.ALTIDX].values,
                    model_dataset[model_dataset.dc.CASEPTR].values,
                    datatree.n_alts,
                ),
                dims=[datatree.CASEID, datatree.ALTID],
                coords={
                    datatree.CASEID: model_dataset.coords[datatree.CASEID],
                    datatree.ALTID: model_dataset.coords[datatree.ALTID],
                },
                name="ch",
            )
            model_dataset = model_dataset.drop_vars(["choice_ce_data"])
            model_dataset["ch"] = da_ch
    if "choice_co_code" in request:
        log.debug(f"requested choice_co_code data: {request['choice_co_code']}")
        if isinstance(datasource, DataTree):
            choicecodes = datasource.get(
                request["choice_co_code"], broadcast=False, coords=False
            )
        else:
            choicecodes = datasource.get(request["choice_co_code"])
        da_ch = DataArray(
            float_dtype(0),
            dims=[datatree.CASEID, datatree.ALTID],
            coords={
                datatree.CASEID: model_dataset.coords[datatree.CASEID],
                datatree.ALTID: model_dataset.coords[datatree.ALTID],
            },
            name="ch",
        )
        for i, a in enumerate(model_dataset.coords[datatree.ALTID]):
            if isinstance(a, DataArray):
                a = a.item()
            da_ch[:, i] = np.asarray(choicecodes == a)
        model_dataset = model_dataset.merge(da_ch)
        if make_unused_flows:
            from .numba_stream import OneHotStreamer

            flows["choice_co_code"] = OneHotStreamer(
                request["choice_co_code"], datatree.root_dataset.coords[datatree.ALTID]
            )
    if "choice_co" in request:
        log.debug(f"requested choice_co_vars data: {request['choice_co']}")
        da_ch = DataArray(
            float_dtype(0),
            dims=[datatree.CASEID, datatree.ALTID],
            coords={
                datatree.CASEID: model_dataset.coords[datatree.CASEID],
                datatree.ALTID: model_dataset.coords[datatree.ALTID],
            },
            name="ch",
        )
        for i, a in enumerate(model_dataset.dc.alts_mapping):
            choice_expr = request["choice_co"][a]
            da_ch[:, i] = datatree_co.dc.get_expr(choice_expr).values
        model_dataset = model_dataset.merge(da_ch)
    if "choice_any" in request:
        log.debug(f"requested choice_any data: {request['choice_any']}")
        raise NotImplementedError("choice_any")

    if "weight_co" in request:
        log.debug(f"requested weight_co data: {request['weight_co']}")
        model_dataset, flows["weight_co"] = _prep_co(
            model_dataset,
            datatree_co,
            [request["weight_co"]],
            tag="wt",
            preserve_vars=make_unused_flows,
            dtype=float_dtype,
            cache_dir=cache_dir,
            flow=flows.get("weight_co"),
        )

    if "group_co" in request:
        log.debug(f"requested group_co data: {request['group_co']}")
        model_dataset, flows["group_co"] = _prep_co(
            model_dataset,
            datatree_co,
            [request["group_co"]],
            tag="group",
            preserve_vars=make_unused_flows,
            dtype=np.int64,
            cache_dir=cache_dir,
            flow=flows.get("group_co"),
        )

    if "avail_ca" in request:
        log.debug(f"requested avail_ca data: {request['avail_ca']}")
        casealt_dim = datatree.root_dataset.attrs.get(_CASEALT)
        if casealt_dim is None:
            model_dataset, flows["avail_ca"] = _prep_ca(
                model_dataset,
                datatree,
                request["avail_ca"],
                tag="av",
                preserve_vars=False,
                dtype=np.int8,
                cache_dir=cache_dir,
                flow=flows.get("avail_ca"),
                force_flow=make_unused_flows,
            )
        else:
            if (
                request["avail_ca"] in {"1", "True", "1.0"}
                and model_dataset.dc.CASEPTR is not None
                and model_dataset.dc.ALTIDX is not None
            ):
                da_av = DataArray(
                    ce_bool_to_dense(
                        model_dataset[model_dataset.dc.ALTIDX].values,
                        model_dataset[model_dataset.dc.CASEPTR].values,
                        datatree.n_alts,
                    ),
                    dims=[datatree.CASEID, datatree.ALTID],
                    coords={
                        datatree.CASEID: model_dataset.coords[datatree.CASEID],
                        datatree.ALTID: model_dataset.coords[datatree.ALTID],
                    },
                    name="av",
                )
            else:
                model_dataset, flows["avail_ce"] = _prep_ce(
                    model_dataset,
                    datatree,
                    request["avail_ca"],
                    v_tag="avail_ca",
                    s_tag="avail_ce",
                    preserve_vars=False,
                    dtype=np.int8,
                    cache_dir=cache_dir,
                    flow=flows.get("avail_ce"),
                    attach_indexes=False,
                )
                da_av = DataArray(
                    ce_to_dense(
                        model_dataset["avail_ce_data"].values,
                        model_dataset[model_dataset.dc.ALTIDX].values,
                        model_dataset[model_dataset.dc.CASEPTR].values,
                        datatree.n_alts,
                    ),
                    dims=[datatree.CASEID, datatree.ALTID],
                    coords={
                        datatree.CASEID: model_dataset.coords[datatree.CASEID],
                        datatree.ALTID: model_dataset.coords[datatree.ALTID],
                    },
                    name="av",
                )
                model_dataset = model_dataset.drop_vars(["avail_ce_data"])
            model_dataset["av"] = da_av
    if "avail_co" in request:
        log.debug(f"requested avail_co data: {request['avail_co']}")
        av_co_expressions = {
            a: request["avail_co"].get(a, "0")
            for a in model_dataset.coords[datatree.ALTID].values
        }
        model_dataset, flows["avail_co"] = _prep_co(
            model_dataset,
            datatree_co,
            av_co_expressions,
            tag="av",
            preserve_vars=make_unused_flows,
            dtype=np.int8,
            dim_name=datatree.ALTID,
            cache_dir=cache_dir,
            flow=flows.get("avail_co"),
        )
    if "avail_any" in request:
        log.debug(f"requested avail_any data: {request['avail_any']}")
        model_dataset = model_dataset.assign(
            {
                "av": DataArray.ones(
                    model_dataset.dc.caseids(),
                    model_dataset.dc.altids(),
                    dtype=np.int8,
                ),
            }
        )

    return model_dataset, flows


def flownamer(tag, definition_spec, extra_hash_features=()):
    import base64
    import hashlib

    defs_hash = hashlib.md5()
    defs_hash.update(str(tag).encode("utf8"))
    for k, v in definition_spec.items():
        defs_hash.update(str(k).encode("utf8"))
        defs_hash.update(str(v).encode("utf8"))
    for k in extra_hash_features:
        defs_hash.update(str(k).encode("utf8"))
    return "pipeline_" + (base64.b32encode(defs_hash.digest())).decode().replace(
        "=", ""
    )


def _prep_ca(
    model_dataset,
    shared_data_ca,
    vars_ca,
    tag="ca",
    preserve_vars=True,
    dtype=None,
    cache_dir=None,
    flow=None,
    force_flow=False,
):
    from ..dataset import DataArray, DataTree

    assert isinstance(shared_data_ca, DataTree)
    if isinstance(vars_ca, str):
        if (
            not preserve_vars
            and vars_ca in shared_data_ca.root_dataset
            and not force_flow
        ):
            proposal = shared_data_ca.root_dataset[vars_ca]
            if (
                shared_data_ca.CASEID in proposal.sizes
                and shared_data_ca.ALTID in proposal.sizes
            ):
                proposal = proposal.drop_vars(list(proposal.coords)).rename(tag)
                return model_dataset.merge(proposal), vars_ca
    if isinstance(vars_ca, str):
        vars_ca = {vars_ca: vars_ca}
    if not isinstance(vars_ca, dict):
        vars_ca = {i: i for i in vars_ca}
    flowname = flownamer(tag, vars_ca, shared_data_ca._hash_features())
    if isinstance(flow, str):
        raise ValueError("expected flow {flow!r}")
    if flow is None or flowname != flow.name:
        flow = shared_data_ca.setup_flow(vars_ca, cache_dir=cache_dir, name=flowname)
    arr = flow.load(
        shared_data_ca,
        dtype=dtype,
    )
    caseid_dim = shared_data_ca.CASEID
    altid_dim = shared_data_ca.ALTID
    if preserve_vars or len(vars_ca) > 1:
        arr = arr.reshape(
            model_dataset.sizes.get(caseid_dim),
            model_dataset.sizes.get(altid_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, altid_dim, f"var_{tag}"],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                altid_dim: model_dataset.coords[altid_dim],
                f"var_{tag}": list(vars_ca.keys()),
            },
            name=tag,
        )
    else:
        arr = arr.reshape(
            model_dataset.sizes.get(caseid_dim),
            model_dataset.sizes.get(altid_dim),
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, altid_dim],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                altid_dim: model_dataset.coords[altid_dim],
            },
            name=tag,
        )
    return model_dataset.merge(da), flow


def _prep_ce(
    model_dataset,
    datatree,
    vars_ca,
    v_tag="ca",
    s_tag="ce",
    preserve_vars=True,
    dtype=None,
    cache_dir=None,
    flow=None,
    attach_indexes=True,
):
    from ..dataset import DataArray, DataTree

    assert isinstance(datatree, DataTree)
    if isinstance(vars_ca, str):
        if not preserve_vars and vars_ca in datatree.root_dataset:
            proposal = datatree.root_dataset[vars_ca]
            if datatree.CASEALT in proposal.sizes:
                proposal = proposal.drop_vars(list(proposal.coords)).rename(
                    f"{s_tag}_data"
                )
                return model_dataset.merge(proposal), flow
    if isinstance(vars_ca, str):
        vars_ca = {vars_ca: vars_ca}
    if not isinstance(vars_ca, dict):
        vars_ca = {i: i for i in vars_ca}
    flowname = flownamer(s_tag, vars_ca, datatree._hash_features())
    if flow is None or flowname != flow.name:
        flow = datatree.setup_flow(vars_ca, cache_dir=cache_dir, name=flowname)
    arr = flow.load(
        datatree,
        dtype=dtype,
    )
    casealt_dim = datatree.CASEALT
    if preserve_vars or len(vars_ca) > 1:
        arr = arr.reshape(
            model_dataset.sizes.get(casealt_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[casealt_dim, f"var_{v_tag}"],
            coords={
                casealt_dim: model_dataset.coords[casealt_dim],
                f"var_{v_tag}": list(vars_ca.keys()),
            },
            name=f"{s_tag}_data",
        )
    else:
        arr = arr.reshape(-1)
        da = DataArray(
            arr,
            dims=[casealt_dim],
            coords={
                casealt_dim: model_dataset.coords[casealt_dim],
            },
            name=f"{s_tag}_data",
        )
    model_dataset = model_dataset.merge(da)

    if attach_indexes:
        altidx = datatree.root_dataset.coords[datatree.ALTIDX]
        altidx = altidx.drop_vars(list(altidx.coords))
        model_dataset[datatree.ALTIDX] = altidx
        model_dataset.dc.ALTIDX = datatree.ALTIDX
        caseptr = datatree.root_dataset[datatree.CASEPTR]
        caseptr = caseptr.drop_vars(list(caseptr.coords))
        model_dataset[datatree.CASEPTR] = caseptr
        model_dataset.dc.CASEPTR = datatree.CASEPTR
        model_dataset = model_dataset.assign_coords(
            {
                datatree.CASEID: DataArray(
                    datatree.caseids(),
                    dims=(datatree.CASEID),
                ),
                datatree.ALTID: DataArray(
                    datatree.altids(),
                    dims=(datatree.ALTID),
                ),
            }
        )
    model_dataset.dc.CASEID = datatree.CASEID
    model_dataset.dc.ALTID = datatree.ALTID
    return model_dataset, flow


def _prep_co(
    model_dataset,
    shared_data_co,
    vars_co,
    tag="co",
    preserve_vars=True,
    dtype=None,
    dim_name=None,
    cache_dir=None,
    flow=None,
):
    from ..dataset import DataArray, DataTree

    assert isinstance(shared_data_co, DataTree)
    if not isinstance(vars_co, dict):
        vars_co = {i: i for i in vars_co}
    flowname = flownamer(tag, vars_co, shared_data_co._hash_features())
    if flow is None or flowname != flow.name:
        flow = shared_data_co.setup_flow(vars_co, cache_dir=cache_dir, name=flowname)
    arr = flow.load(
        shared_data_co,
        dtype=dtype,
    )
    caseid_dim = shared_data_co.CASEID
    if preserve_vars or len(vars_co) > 1:
        if dim_name is None:
            dim_name = f"var_{tag}"
        arr = arr.reshape(
            model_dataset.sizes.get(caseid_dim),
            -1,
        )
        da = DataArray(
            arr,
            dims=[caseid_dim, dim_name],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
                dim_name: list(vars_co.keys()),
            },
            name=tag,
        )
    else:
        arr = arr.reshape(
            model_dataset.sizes.get(caseid_dim),
        )
        da = DataArray(
            arr,
            dims=[caseid_dim],
            coords={
                caseid_dim: model_dataset.coords[caseid_dim],
            },
            name=tag,
        )
    return model_dataset.merge(da), flow


@nb.njit
def ce_to_dense(ce_data, ce_altidx, ce_caseptr, n_alts):
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:, -1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0], n_alts, *ce_data.shape[1:])
    out = np.zeros(shape, dtype=ce_data.dtype)
    c = 0
    for row in range(ce_data.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
        a = ce_altidx[row]
        out[c, a, ...] = ce_data[row, ...]
    return out


@nb.njit
def ce_bool_to_dense(ce_altidx, ce_caseptr, n_alts):
    if ce_caseptr.ndim == 2:
        ce_caseptr1 = ce_caseptr[:, -1]
    else:
        ce_caseptr1 = ce_caseptr[1:]
    shape = (ce_caseptr1.shape[0], n_alts)
    out = np.zeros(shape, dtype=np.int8)
    c = 0
    for row in range(ce_altidx.shape[0]):
        if row == ce_caseptr1[c]:
            c += 1
        a = ce_altidx[row]
        out[c, a, ...] = 1
    return out




# Example code extracted automatically from documentation.
# Do not edit these files manually.
from ._030_mnl_constrain import example as _030_mnl_constrain
from ._001_mnl import example as _001_mnl
from ._017_mnl import example as _017_mnl
from ._022_nl import example as _022_nl
from ._151_swissmetro_panel_latent_class import example as _151_swissmetro_panel_latent_class
from ._102_swissmetro_weighted import example as _102_swissmetro_weighted
from ._109_swissmetro_nl import example as _109_swissmetro_nl
from ._101_swissmetro_mnl import example as _101_swissmetro_mnl
from ._204_exville_dest_only import example as _204_exville_dest_only
from ._251_bySpec import example as _251_bySpec
from ._200_exampville import example as _200_exampville
from ._203_exville_dest import example as _203_exville_dest
from ._201_exville_mode_choice import example as _201_exville_mode_choice
from ._202_exville_mc_logsums import example as _202_exville_mc_logsums
from ._xlogit_test import example as _xlogit_test

ex = {}
ex[30] = _030_mnl_constrain
ex[1] = _001_mnl
ex[17] = _017_mnl
ex[22] = _022_nl
ex[151] = _151_swissmetro_panel_latent_class
ex[102] = _102_swissmetro_weighted
ex[109] = _109_swissmetro_nl
ex[101] = _101_swissmetro_mnl
ex[204] = _204_exville_dest_only
ex[251] = _251_bySpec
ex[200] = _200_exampville
ex[203] = _203_exville_dest
ex[201] = _201_exville_mode_choice
ex[202] = _202_exville_mc_logsums
ex['xlogit'] = _xlogit_test
